#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from xml.etree import ElementTree

from oslo_config import cfg
from oslo_log import log as logging
import stevedore

from networking_baremetal import constants
from networking_baremetal import exceptions

DRIVER_NAMESPACE = 'networking_baremetal.drivers'

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


def txt_subelement(parent, tag, text, *args, **kwargs):
    element = ElementTree.SubElement(parent, tag, *args, **kwargs)
    element.text = text
    return element


def config_to_xml(config):
    element = ElementTree.Element(constants.CFG_ELEMENT)
    for conf in config:
        element.append(conf.to_xml_element())
    return ElementTree.tostring(element).decode("utf-8")


def driver_mgr(device_id):
    driver = CONF[device_id].driver
    try:
        mgr = stevedore.driver.DriverManager(
            namespace=DRIVER_NAMESPACE,
            name=driver,
            invoke_on_load=True,
            invoke_args=(device_id,),
            on_load_failure_callback=_load_failure_hook
        )
    except stevedore.exception.NoUniqueMatch as exc:
        raise exceptions.DriverEntrypointLoadError(
            entry_point=f'{DRIVER_NAMESPACE}.{driver}',
            err=exc)

    return mgr.driver


def _load_failure_hook(manager, entrypoint, exception):
    LOG.error("Driver manager %(manager)s failed to load device plugin "
              "%(entrypoint)s: %(exp)s",
              {'manager': manager, 'entrypoint': entrypoint, 'exp': exception})
    raise exceptions.DriverEntrypointLoadError(entry_point=entrypoint,
                                               err=exception)



# Copyright 2017 Cisco Systems, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import socket
import sys
from urllib import parse as urlparse

from neutron.agent import rpc as agent_rpc
from neutron.common import config as common_config
from neutron.conf.agent import common as agent_config
from neutron_lib.agent import topics
from neutron_lib import constants as n_const
from neutron_lib import context
from openstack import exceptions as sdk_exc
from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging
from oslo_service import loopingcall
from oslo_service import service
from oslo_utils import timeutils
from oslo_utils import uuidutils
from tooz import hashring

from networking_baremetal import constants
from networking_baremetal import ironic_client

CONF = cfg.CONF
LOG = logging.getLogger(__name__)
CONF.import_group('AGENT', 'neutron.plugins.ml2.drivers.agent.config')


def list_opts():
    return [('agent', agent_config.AGENT_STATE_OPTS)]


def _get_notification_transport_url():
    url = urlparse.urlparse(CONF.transport_url)
    if (CONF.oslo_messaging_rabbit.amqp_auto_delete is False
            and not getattr(CONF.oslo_messaging_rabbit, 'rabbit_quorum_queue',
                            None)):
        q = urlparse.parse_qs(url.query)
        q.update({'amqp_auto_delete': ['true']})
        query = urlparse.urlencode({k: v[0] for k, v in q.items()})
        url = url._replace(query=query)
    return urlparse.urlunparse(url)


def _set_up_notifier(transport, uuid):
    return oslo_messaging.Notifier(
        transport,
        publisher_id='ironic-neutron-agent-' + uuid,
        driver='messagingv2',
        topics=['ironic-neutron-agent-member-manager'])


def _set_up_listener(transport, agent_id):
    targets = [
        oslo_messaging.Target(topic='ironic-neutron-agent-member-manager')]
    endpoints = [HashRingMemberManagerNotificationEndpoint()]
    return oslo_messaging.get_notification_listener(
        transport, targets, endpoints, pool=agent_id)


class HashRingMemberManagerNotificationEndpoint(object):
    """Class variables members and hashring is shared by all instances"""

    filter_rule = oslo_messaging.NotificationFilter(
        publisher_id='^ironic-neutron-agent.*')

    members = []
    hashring = hashring.HashRing([])

    def info(self, ctxt, publisher_id, event_type, payload, metadata):

        timestamp = timeutils.utcnow_ts()
        # Add members or update timestamp for existing members
        if payload['id'] not in [x['id'] for x in self.members]:
            try:
                LOG.info('Adding member id %s on host %s to hashring.',
                         payload['id'], payload['host'])
                self.hashring.add_node(payload['id'])
                self.members.append(payload)
            except Exception:
                LOG.exception('Failed to add member %s to hash ring!',
                              payload['id'])
        else:
            for member in self.members:
                if payload['id'] == member['id']:
                    member['timestamp'] = payload['timestamp']

        # Remove members that have not checked in for a while
        for member in self.members:
            if (timestamp - member['timestamp']) > (
                    CONF.AGENT.report_interval * 3):
                try:
                    LOG.info('Removing member %s on host %s from hashring.',
                             member['id'], member['host'])
                    self.hashring.remove_node(member['id'])
                    self.members.remove(member)
                except Exception:
                    LOG.exception('Failed to remove member %s from hash ring!',
                                  member['id'])

        return oslo_messaging.NotificationResult.HANDLED


class BaremetalNeutronAgent(service.ServiceBase):

    def __init__(self):
        self.context = context.get_admin_context_without_session()
        self.agent_id = uuidutils.generate_uuid(dashed=True)
        self.agent_host = socket.gethostname()
        self.heartbeat = None
        self.notify_agents = None

        # Set up oslo_messaging notifier and listener to keep track of other
        # members
        # NOTE(hjensas): Override the control_exchange for the notification
        # transport to allow setting amqp_auto_delete = true.
        # TODO(hjensas): Remove this and override the exchange when setting up
        # the notifier once the fix for bug is available.
        #   https://bugs.launchpad.net/oslo.messaging/+bug/1814797
        CONF.set_override('control_exchange', 'ironic-neutron-agent')
        self.transport = oslo_messaging.get_notification_transport(
            CONF, url=_get_notification_transport_url())
        self.notifier = _set_up_notifier(self.transport, self.agent_id)
        # Note(hjensas): We need to have listener consuming the non-pool queue.
        # See bug: https://bugs.launchpad.net/oslo.messaging/+bug/1814544
        self.listener = _set_up_listener(self.transport, None)
        self.pool_listener = _set_up_listener(self.transport, '-'.join(
            ['ironic-neutron-agent-member-manager-pool', self.agent_id]))

        self.member_manager = HashRingMemberManagerNotificationEndpoint()

        self.state_rpc = agent_rpc.PluginReportStateAPI(topics.REPORTS)
        self.ironic_client = ironic_client.get_client()
        self.reported_nodes = {}
        LOG.info('Agent networking-baremetal initialized.')

    def start(self):
        LOG.info('Starting agent networking-baremetal.')
        self.pool_listener.start()
        self.listener.start()
        self.notify_agents = loopingcall.FixedIntervalLoopingCall(
            self._notify_peer_agents)
        self.notify_agents.start(interval=(CONF.AGENT.report_interval / 3))
        self.heartbeat = loopingcall.FixedIntervalLoopingCall(
            self._report_state)
        self.heartbeat.start(interval=CONF.AGENT.report_interval,
                             initial_delay=CONF.AGENT.report_interval)
        self.cleanup_stale_agents()

    def stop(self, failure=False):
        LOG.info('Stopping agent networking-baremetal.')
        if self.heartbeat:
            self.heartbeat.stop()
        if self.notify_agents:
            self.notify_agents.stop()
        self.listener.stop()
        self.pool_listener.stop()
        self.listener.wait()
        self.pool_listener.wait()
        if failure:
            # This will generate a SIGABORT for the process which forces it
            # to exit, which seems cleaner to force the process to exit
            # than os.exit and avoids threading constraints.
            os.abort()

    def reset(self):
        LOG.info('Resetting agent networking-baremetal.')
        if self.heartbeat:
            self.heartbeat.stop()
        if self.notify_agents:
            self.notify_agents.stop()
        self.listener.stop()
        self.pool_listener.stop()
        self.listener.wait()
        self.pool_listener.wait()

    def wait(self):
        pass

    def _notify_peer_agents(self):
        try:
            self.notifier.info({
                'ironic-neutron-agent': 'heartbeat'},
                'ironic-neutron-agent-member-manager',
                {'id': self.agent_id,
                 'host': self.agent_host,
                 'timestamp': timeutils.utcnow_ts()})
        except Exception:
            LOG.exception('Failed to send hash ring membership heartbeat!')

    def get_template_node_state(self, node_uuid):
        return {
            'binary': constants.BAREMETAL_BINARY,
            'host': node_uuid,
            'topic': n_const.L2_AGENT_TOPIC,
            'configurations': {
                'bridge_mappings': {},
                'log_agent_heartbeats': CONF.AGENT.log_agent_heartbeats,
            },
            'start_flag': False,
            'agent_type': constants.BAREMETAL_AGENT_TYPE,
            'action': 'update'}

    def _report_state(self):
        node_states = {}
        ironic_ports = self.ironic_client.ports(details=True)

        # NOTE: the above calls returns a generator, so we need to handle
        # exceptions that happen just before the first loop iteration, when
        # the actual request to ironic happens
        try:
            for port in ironic_ports:
                node = port.node_id
                if (self.agent_id not in
                        self.member_manager.hashring[node.encode('utf-8')]):
                    continue
                template_node_state = self.get_template_node_state(node)
                node_states.setdefault(node, template_node_state)
                mapping = node_states[
                    node]["configurations"]["bridge_mappings"]
                if port.physical_network is not None:
                    mapping[port.physical_network] = "yes"
        except sdk_exc.OpenStackCloudException:
            LOG.exception("Failed to get ironic ports data! "
                          "Not reporting state.")
            try:
                # Replace the client, just to be on the safe side in
                # the event there was some sort of hard/breaking failure.
                self.ironic_client = ironic_client.get_client()
            except Exception:
                # Failed to re-launch a new client, aborting.
                self.stop(failure=True)
            return
        abort_operation = False
        for state in node_states.values():
            # If the node was not previously reported with current
            # configuration set the start_flag True.
            # NOTE(TheJulia) reported_nodes is an internal list of nodes
            # we *have* updated.
            if not state['configurations'] == self.reported_nodes.get(
                    state['host']):
                state.update({'start_flag': True})
                LOG.info('Reporting state for host agent %s with new '
                         'configuration: %s',
                         state['host'], state['configurations'])
            try:
                LOG.debug('Reporting state for host: %s with configuration: '
                          '%s', state['host'], state['configurations'])
                self.state_rpc.report_state(self.context, state)
            except AttributeError:
                # This means the server does not support report_state
                LOG.exception("Neutron server does not support state report. "
                              "State report for this agent will be disabled.")
                # Don't continue reporting the remaining agents in this case.
                abort_operation = True
                break
            except Exception:
                LOG.exception("Failed reporting state!")
                # Don't continue reporting the remaining nodes if one failed.
                return
            self.reported_nodes.update(
                {state['host']: state['configurations']})

        # Identify nodes that are no longer present in Ironic by subtracting
        # the keys of `node_states` from the keys of `reported_nodes`. Then
        # delete agents for nodes that are no longer present.
        deleted_nodes = self.reported_nodes.keys() - node_states.keys()
        deleted_agents = self._delete_agents(deleted_nodes)
        for node in deleted_agents:
            self.reported_nodes.pop(node)

        if abort_operation:
            # We don't expect the agent to work, and as such we should call
            # stop so the program unwinds and begins to exit.
            self.stop(failure=True)

    def _get_down_agents(self):
        """Retrieves a list of inactive Baremetal agents.

        Fetch a list of inactive Baremetal agents. It interacts with
        the state_rpc object to call the 'get_agents' method, which
        retrieves agents based on the provided parameters.

        :returns: (list) Inactive Baremetal agents.
        """
        down_bm_agents = []
        try:
            down_bm_agents = self.state_rpc.get_agents(
                self.context,
                agent_type=constants.BAREMETAL_AGENT_TYPE,
                is_active=False)
        except oslo_messaging.NoSuchMethod:
            LOG.warning("Neutron server doesn't support "
                        "`get_agents` endpoint.")

        return down_bm_agents

    def _get_nodes_not_found(self, down_bm_agents):
        """Identifies nodes that are not found in the Ironic

        The method iterates over each agent in the 'down_bm_agents' list.
        For each agent, it attempts to retrieve the corresponding node using
        the Ironic client's 'get_node' method. If the node is not found the
        node is appended to the 'nodes_not_found' list.

        :param down_bm_agents: (list) Agents that are down in Neutron.
        :return: (list) Nodes that are not found in Ironic.
        """
        nodes_not_found = []
        for agent in down_bm_agents:
            node = agent['host']
            try:
                self.ironic_client.get_node(node)
            except sdk_exc.NotFoundException:
                nodes_not_found.append(node)

        return nodes_not_found

    def _delete_agents(self, nodes, log=True):
        """Delete agents for nodes that are not found in ironic

        Clean up agent records in neutron for ironic nodes that have been
        removed from the system.

        :param nodes_not_found: (list) Nodes that are not found in Ironic.
        :log: (bool) Log the actions taken.
        :return: (list) Agents that have been deleted in Neutron.
        """
        deleted_agents = []
        for node in nodes:
            if log:
                LOG.info('Removing agent for host: %s', node)
            try:
                kwargs = {'host': node,
                          'agent_type': constants.BAREMETAL_AGENT_TYPE}
                self.state_rpc.delete_agent(self.context, **kwargs)
                deleted_agents.append(node)
            except oslo_messaging.NoSuchMethod:
                LOG.warning("Neutron server doesn't support "
                            "`delete_agent` endpoint.")
                break

        return deleted_agents

    def cleanup_stale_agents(self):
        """Cleans up stale baremetal agents

        This method identifies baremetal agents that are marked as
        inactive in the Neutron server and are not associated with
        any nodes in Ironic. It then deletes these stale agents.
        """
        down_bm_agents = self._get_down_agents()
        nodes_not_found = self._get_nodes_not_found(down_bm_agents)
        deleted_agents = self._delete_agents(nodes_not_found, log=False)

        if deleted_agents:
            LOG.info("Stale baremetal agent for hosts was removed: %s",
                     ", ".join(deleted_agents))


def _unregiser_deprecated_opts():
    CONF.reset()
    CONF.unregister_opts(
        [CONF._groups[ironic_client.IRONIC_GROUP]._opts[opt]['opt']
         for opt in ironic_client._deprecated_opts],
        group=ironic_client.IRONIC_GROUP)


def main():
    common_config.register_common_config_options()
    # TODO(hjensas): Imports from neutron in ironic_neutron_agent registers the
    # client options. We need to unregister the options we are deprecating
    # first to avoid DuplicateOptError. Remove this when dropping deprecations.
    _unregiser_deprecated_opts()
    common_config.init(sys.argv[1:])
    common_config.setup_logging()
    agent = BaremetalNeutronAgent()
    launcher = service.launch(cfg.CONF, agent, restart_method='mutate')
    launcher.wait()



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import abc


class BaseDeviceClient(object, metaclass=abc.ABCMeta):

    def __init__(self, device):
        self.device = device

    def get_client_args(self):
        """Get client connection arguments from configuration"""

    def get(self, **kwargs):
        """Get current configuration/state from device"""

    def edit_config(self, config):
        """Edit configuration on the device

        :param config: The configuration to apply to the device
        """


class BaseDeviceDriver(object, metaclass=abc.ABCMeta):

    SUPPORTED_BOND_MODES = set()

    def __init__(self, device):
        self.client = BaseDeviceClient(device)
        self.device = device

    def load_config(self):
        """Register driver specific configuration

        All drivers should register driver specific options in the
        device specific config group. This method will be called
        during mechanism driver initialization.
        """

    def validate(self):
        """Driver validation

        This method will be called during mechanism driver
        initialization. Raising any exception other than
        DriverValidationError will cause service initialization
        failure.

        :raises DriverValidationError: On validation failure.
        """

    def create_network(self, context):
        """Create network on device

        :param context: NetworkContext instance describing the new
            network.
        """

    def update_network(self, context):
        """Update network on device

        :param context: NetworkContext instance describing the new
            network.
        """

    def delete_network(self, context):
        """Delete network on device

        :param context: NetworkContext instance describing the new
            network.
        """

    def create_port(self, context, segment, links):
        """Create/Configure port on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param segment: segment dictionary describing segment to bind
        :param links: Local link information filtered for the device.
        """

    def update_port(self, context, links):
        """Update port on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        """

    def delete_port(self, context, links, current=True):
        """Delete/Un-configure port on device

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param links: Local link information filtered for the device.
        :param current: Boolean, when true use context.current, when
            false use context.original
        """



# Copyright 2017 Red Hat Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from urllib import parse as urlparse

from neutron.agent import rpc as agent_rpc
from neutron.tests import base
from neutron_lib import constants as n_const
from openstack import connection
from openstack import exceptions as sdk_exc
from oslo_config import fixture as config_fixture
from tooz import hashring

from networking_baremetal.agent import ironic_neutron_agent
from networking_baremetal import constants
from networking_baremetal import ironic_client


class FakePort1(object):
    def __init__(self, physnet='physnet1'):
        self.uuid = '11111111-2222-3333-4444-555555555555'
        self.node_id = '55555555-4444-3333-2222-111111111111'
        self.physical_network = physnet


class FakePort2(object):
    def __init__(self, physnet='physnet2'):
        self.uuid = '11111111-aaaa-3333-4444-555555555555'
        self.node_id = '55555555-4444-3333-aaaa-111111111111'
        self.physical_network = physnet


@mock.patch.object(ironic_client, '_get_ironic_session', autospec=True)
@mock.patch.object(connection.Connection, 'baremetal', autospec=True)
class TestBaremetalNeutronAgent(base.BaseTestCase):
    def setUp(self):
        super(TestBaremetalNeutronAgent, self).setUp()
        self.context = object()
        self.conf = self.useFixture(config_fixture.Config())
        self.conf.config(transport_url='rabbit://user:password@host/')

    def test_get_template_node_state(self, mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        # Verify agent binary
        expected = constants.BAREMETAL_BINARY
        self.assertEqual(expected,
                         self.agent.get_template_node_state(
                             'uuid')['binary'])

        # Verify agent_type is Baremetal Node
        expected = constants.BAREMETAL_AGENT_TYPE
        self.assertEqual(expected,
                         self.agent.get_template_node_state(
                             'uuid')['agent_type'])
        # Verify topic
        expected = n_const.L2_AGENT_TOPIC
        self.assertEqual(expected,
                         self.agent.get_template_node_state(
                             'uuid')['topic'])
        # Verify host
        expected = 'the_node_uuid'
        self.assertEqual(expected,
                         self.agent.get_template_node_state(
                             'the_node_uuid')['host'])

    def test_report_state_one_node_one_port(self, mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        with mock.patch.object(self.agent.state_rpc, 'report_state',
                               autospec=True) as mock_report_state:
            self.agent.ironic_client = mock_conn
            mock_conn.ports.return_value = iter([FakePort1()])
            self.agent.agent_id = 'agent_id'
            self.agent.member_manager.hashring = hashring.HashRing(
                [self.agent.agent_id])

            expected = {
                'topic': n_const.L2_AGENT_TOPIC,
                'start_flag': True,
                'binary': constants.BAREMETAL_BINARY,
                'host': '55555555-4444-3333-2222-111111111111',
                'configurations': {
                    'bridge_mappings': {
                        'physnet1': 'yes'
                    },
                    'log_agent_heartbeats': False,
                },
                'agent_type': constants.BAREMETAL_AGENT_TYPE,
                'action': 'update'
            }
            self.agent._report_state()
            mock_report_state.assert_called_with(self.agent.context, expected)

    def test_report_state_with_log_agent_heartbeats(self, mock_conn,
                                                    mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        with mock.patch.object(self.agent.state_rpc, 'report_state',
                               autospec=True) as mock_report_state:
            self.conf.config(log_agent_heartbeats=True, group='AGENT')
            self.agent.ironic_client = mock_conn
            mock_conn.ports.return_value = iter([FakePort1()])
            self.agent.agent_id = 'agent_id'
            self.agent.member_manager.hashring = hashring.HashRing(
                [self.agent.agent_id])

            expected = {
                'topic': n_const.L2_AGENT_TOPIC,
                'start_flag': True,
                'binary': constants.BAREMETAL_BINARY,
                'host': '55555555-4444-3333-2222-111111111111',
                'configurations': {
                    'bridge_mappings': {
                        'physnet1': 'yes'
                    },
                    'log_agent_heartbeats': True,
                },
                'agent_type': constants.BAREMETAL_AGENT_TYPE,
                'action': 'update'
            }

            self.agent._report_state()
            mock_report_state.assert_called_with(self.agent.context, expected)

    def test_start_flag_false_on_update_no_config_change(self, mock_conn,
                                                         mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        with mock.patch.object(self.agent.state_rpc, 'report_state',
                               autospec=True) as mock_report_state:
            self.agent.ironic_client = mock_conn
            mock_conn.ports.return_value = iter([FakePort1()])
            self.agent.agent_id = 'agent_id'
            self.agent.member_manager.hashring = hashring.HashRing(
                [self.agent.agent_id])

            expected = {
                'topic': n_const.L2_AGENT_TOPIC,
                'start_flag': 'PLACEHOLDER',
                'binary': constants.BAREMETAL_BINARY,
                'host': '55555555-4444-3333-2222-111111111111',
                'configurations': {
                    'bridge_mappings': {
                        'physnet1': 'yes'
                    },
                    'log_agent_heartbeats': False,
                },
                'agent_type': constants.BAREMETAL_AGENT_TYPE,
                'action': 'update'
            }

            # First time report start_flag is True
            expected.update({'start_flag': True})
            self.agent._report_state()
            mock_report_state.assert_called_with(self.agent.context, expected)
            # Subsequent times report start_flag is False
            mock_conn.ports.return_value = iter([FakePort1()])
            expected.update({'start_flag': False})
            self.agent._report_state()
            mock_report_state.assert_called_with(self.agent.context, expected)

    def test_start_flag_true_on_update_after_config_change(self, mock_conn,
                                                           mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        with mock.patch.object(self.agent.state_rpc, 'report_state',
                               autospec=True) as mock_report_state:
            self.agent.ironic_client = mock_conn
            mock_conn.ports.return_value = iter([FakePort1()])
            self.agent.agent_id = 'agent_id'
            self.agent.member_manager.hashring = hashring.HashRing(
                [self.agent.agent_id])

            expected = {
                'topic': n_const.L2_AGENT_TOPIC,
                'start_flag': 'PLACEHOLDER',
                'binary': constants.BAREMETAL_BINARY,
                'host': '55555555-4444-3333-2222-111111111111',
                'configurations': {
                    'bridge_mappings': {
                        'physnet1': 'yes'
                    },
                    'log_agent_heartbeats': False,
                },
                'agent_type': constants.BAREMETAL_AGENT_TYPE,
                'action': 'update'
            }

            # First time report start_flag is True
            expected.update({'start_flag': True})
            self.agent._report_state()
            mock_report_state.assert_called_with(self.agent.context, expected)
            # Subsequent times report start_flag is False
            mock_conn.ports.return_value = iter([FakePort1()])
            expected.update({'start_flag': False})
            self.agent._report_state()
            mock_report_state.assert_called_with(self.agent.context, expected)
            # After bridge_mapping config change start_flag is True once
            mock_conn.ports.return_value = iter(
                [FakePort1(physnet='new_physnet')])
            expected.update({'configurations': {
                'bridge_mappings': {'new_physnet': 'yes'},
                'log_agent_heartbeats': False}})
            expected.update({'start_flag': True})
            self.agent._report_state()
            mock_report_state.assert_called_with(self.agent.context, expected)
            # Subsequent times report start_flag is False
            mock_conn.ports.return_value = iter(
                [FakePort1(physnet='new_physnet')])
            expected.update({'start_flag': False})
            self.agent._report_state()
            mock_report_state.assert_called_with(self.agent.context, expected)

    def test_report_state_two_nodes_two_ports(self, mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        with mock.patch.object(self.agent.state_rpc, 'report_state',
                               autospec=True) as mock_report_state:
            self.agent.ironic_client = mock_conn
            mock_conn.ports.return_value = iter([FakePort1(), FakePort2()])
            self.agent.agent_id = 'agent_id'
            self.agent.member_manager.hashring = hashring.HashRing(
                [self.agent.agent_id])

            expected1 = {
                'topic': n_const.L2_AGENT_TOPIC,
                'start_flag': True,
                'binary': constants.BAREMETAL_BINARY,
                'host': '55555555-4444-3333-2222-111111111111',
                'configurations': {
                    'bridge_mappings': {
                        'physnet1': 'yes'
                    },
                    'log_agent_heartbeats': False,
                },
                'agent_type': constants.BAREMETAL_AGENT_TYPE,
                'action': 'update'
            }
            expected2 = {
                'topic': n_const.L2_AGENT_TOPIC,
                'start_flag': True,
                'binary': constants.BAREMETAL_BINARY,
                'host': '55555555-4444-3333-aaaa-111111111111',
                'configurations': {
                    'bridge_mappings': {
                        'physnet2': 'yes'
                    },
                    'log_agent_heartbeats': False,
                },
                'agent_type': constants.BAREMETAL_AGENT_TYPE,
                'action': 'update'
            }

            self.agent._report_state()
            mock_report_state.assert_has_calls(
                [mock.call(self.agent.context, expected1),
                 mock.call(self.agent.context, expected2)], any_order=True)

    def test_report_state_deleted_node(self, mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        with (mock.patch.object(self.agent.state_rpc, 'report_state',
                                autospec=True) as mock_report_state,
              mock.patch.object(self.agent.state_rpc, 'delete_agent',
                                autospec=True) as mock_delete_agent):
            self.agent.ironic_client = mock_conn
            mock_conn.ports.return_value = iter([FakePort1()])
            self.agent.agent_id = 'agent_id'
            self.agent.member_manager.hashring = hashring.HashRing(
                [self.agent.agent_id])

            expected1 = {
                'topic': n_const.L2_AGENT_TOPIC,
                'start_flag': True,
                'binary': constants.BAREMETAL_BINARY,
                'host': '55555555-4444-3333-2222-111111111111',
                'configurations': {
                    'bridge_mappings': {
                        'physnet1': 'yes'
                    },
                    'log_agent_heartbeats': False,
                },
                'agent_type': constants.BAREMETAL_AGENT_TYPE,
                'action': 'update'
            }

            self.agent.reported_nodes = {
                '55555555-4444-3333-2222-111111111111': {},
                '55555555-4444-3333-aaaa-111111111111': {}}
            self.agent._report_state()
            mock_report_state.assert_has_calls(
                [mock.call(self.agent.context, expected1)], any_order=True)
            mock_delete_agent.assert_has_calls(
                [mock.call(self.agent.context,
                           host='55555555-4444-3333-aaaa-111111111111',
                           agent_type=constants.BAREMETAL_AGENT_TYPE)],
                any_order=True)

    @mock.patch.object(ironic_client, 'get_client', autospec=True)
    @mock.patch.object(ironic_neutron_agent.LOG, 'exception', autospec=True)
    def test_ironic_port_list_fail(self, mock_log, mock_get_client,
                                   mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        self.agent.ironic_client = mock_conn

        def mock_generator(details=None):
            raise sdk_exc.OpenStackCloudException()
            yield

        mock_conn.ports.side_effect = mock_generator
        self.agent._report_state()
        self.assertEqual(1, mock_log.call_count)
        # Test initialization triggers the client call once
        # before _report_state is triggered, hence call
        # count below of 2.
        self.assertEqual(2, mock_get_client.call_count)

    @mock.patch.object(ironic_neutron_agent.BaremetalNeutronAgent, 'stop',
                       autospec=True)
    @mock.patch.object(ironic_client, 'get_client', autospec=True)
    @mock.patch.object(ironic_neutron_agent.LOG, 'exception', autospec=True)
    def test_ironic_port_list_fail_breakage(self, mock_log, mock_get_client,
                                            mock_stop, mock_conn,
                                            mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        self.agent.ironic_client = mock_conn
        mock_get_client.side_effect = Exception

        def mock_generator(details=None):
            raise sdk_exc.OpenStackCloudException()
            yield

        mock_conn.ports.side_effect = mock_generator
        self.agent._report_state()
        self.assertEqual(1, mock_log.call_count)
        # Checking the count on stop to see if it is called, as
        # opposed to the get_client method as it is the exception
        # root cause.
        mock_stop.assert_called_once_with(mock.ANY, failure=True)

    @mock.patch.object(ironic_neutron_agent.LOG, 'exception', autospec=True)
    @mock.patch.object(agent_rpc, 'PluginReportStateAPI', autospec=True)
    def test_state_rpc_report_state_fail(self, mock_report_state, mock_log,
                                         mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        self.agent.agent_id = 'agent_id'
        self.agent.member_manager.hashring = hashring.HashRing(
            [self.agent.agent_id])

        self.agent.ironic_client = mock_conn
        self.agent.state_rpc = mock_report_state
        mock_conn.ports.return_value = iter([FakePort1(), FakePort2()])
        mock_report_state.report_state.side_effect = Exception()
        self.agent._report_state()
        self.assertEqual(1, mock_log.call_count)

    @mock.patch.object(ironic_neutron_agent.BaremetalNeutronAgent, 'stop',
                       autospec=True)
    @mock.patch.object(ironic_neutron_agent.LOG, 'exception', autospec=True)
    @mock.patch.object(agent_rpc, 'PluginReportStateAPI', autospec=True)
    def test_state_rpc_report_state_fail_attribute(self, mock_report_state,
                                                   mock_log, mock_stop,
                                                   mock_conn,
                                                   mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        self.agent.agent_id = 'agent_id'
        self.agent.member_manager.hashring = hashring.HashRing(
            [self.agent.agent_id])

        self.agent.ironic_client = mock_conn
        self.agent.state_rpc = mock_report_state
        mock_conn.ports.return_value = iter([FakePort1(), FakePort2()])
        mock_report_state.report_state.side_effect = AttributeError()
        self.agent._report_state()
        self.assertEqual(1, mock_log.call_count)
        mock_stop.assert_called_once_with(mock.ANY, failure=True)

    def test__get_notification_transport_url(self, mock_conn, mock_ir_client):
        self.assertEqual(
            'rabbit://user:password@host/?amqp_auto_delete=true',
            ironic_neutron_agent._get_notification_transport_url())

        self.conf.config(transport_url='rabbit://user:password@host:5672/')
        self.assertEqual(
            'rabbit://user:password@host:5672/?amqp_auto_delete=true',
            ironic_neutron_agent._get_notification_transport_url())

        self.conf.config(transport_url='rabbit://host:5672/')
        self.assertEqual(
            'rabbit://host:5672/?amqp_auto_delete=true',
            ironic_neutron_agent._get_notification_transport_url())

        self.conf.config(transport_url='rabbit://user:password@host/vhost')
        self.assertEqual(
            'rabbit://user:password@host/vhost?amqp_auto_delete=true',
            ironic_neutron_agent._get_notification_transport_url())

        self.conf.config(
            transport_url='rabbit://user:password@host/vhost?foo=bar')
        self.assertEqual(
            # NOTE(hjensas): Parse the url's when comparing, different versions
            # may sort the query different.
            urlparse.urlparse('rabbit://user:password@host/'
                              'vhost?foo=bar&amqp_auto_delete=true'),
            urlparse.urlparse(
                ironic_neutron_agent._get_notification_transport_url()))

        self.conf.config(
            transport_url=('rabbit://user:password@host/vhost?foo=bar&'
                           'amqp_auto_delete=false'))
        self.assertEqual(
            # NOTE(hjensas): Parse the url's when comparing, different versions
            # may sort the query different.
            urlparse.urlparse('rabbit://user:password@host'
                              '/vhost?foo=bar&amqp_auto_delete=true'),
            urlparse.urlparse(
                ironic_neutron_agent._get_notification_transport_url()))

    def test__get_notification_transport_url_auto_delete_enabled(
            self, mock_conn, mock_ir_client):
        self.conf.config(amqp_auto_delete=True, group='oslo_messaging_rabbit')
        self.assertEqual(
            'rabbit://user:password@host/',
            ironic_neutron_agent._get_notification_transport_url())

    def test__get_down_agents(self, mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()

        with mock.patch.object(self.agent.state_rpc, 'get_agents',
                               return_value=[], autospec=True):
            self.assertEqual([], self.agent._get_down_agents())

        with mock.patch.object(self.agent.state_rpc, 'get_agents',
                               return_value=[{'host': 'deleted_host'}],
                               autospec=True):
            self.assertEqual(
                [{'host': 'deleted_host'}], self.agent._get_down_agents())

    def test__get_nodes_not_found(self, mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        self.agent.ironic_client = mock_conn

        down_agents = [{'host': 'deleted_host'},
                       {'host': 'existing_host'}]
        expected = ['deleted_host']
        mock_conn.get_node.side_effect = [
            sdk_exc.NotFoundException(), mock.Mock()]
        self.assertEqual(expected,
                         self.agent._get_nodes_not_found(down_agents))

    def test__delete_agents(self, mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        self.agent.agent_id = 'agent_id'
        self.agent.member_manager.hashring = hashring.HashRing(
            [self.agent.agent_id])

        nodes_not_found = ['host0', 'host1']
        calls = [
            mock.call(self.agent.context,
                      host=nodes_not_found[0],
                      agent_type=constants.BAREMETAL_AGENT_TYPE),
            mock.call(self.agent.context,
                      host=nodes_not_found[1],
                      agent_type=constants.BAREMETAL_AGENT_TYPE)
        ]
        with mock.patch.object(self.agent.state_rpc, 'delete_agent',
                               autospec=True) as mock_delete_agent:
            self.agent._delete_agents(nodes_not_found)
            mock_delete_agent.assert_has_calls(calls, any_order=True)

    def test_cleanup_stale_agents(self, mock_conn, mock_ir_client):
        self.agent = ironic_neutron_agent.BaremetalNeutronAgent()
        with (mock.patch.object(self.agent.state_rpc, 'delete_agent',
                                autospec=True) as mock_delete_agent,
              mock.patch.object(self.agent.state_rpc, 'get_agents',
                                return_value=[{'host': 'deleted_host'}],
                                autospec=True)):
            self.agent.ironic_client = mock_conn
            mock_conn.get_node.side_effect = sdk_exc.NotFoundException()
            self.agent.agent_id = 'agent_id'
            self.agent.member_manager.hashring = hashring.HashRing(
                [self.agent.agent_id])

            self.agent.cleanup_stale_agents()
            kwargs = {'host': 'deleted_host',
                      'agent_type': constants.BAREMETAL_AGENT_TYPE}
            mock_delete_agent.assert_called_with(self.agent.context, **kwargs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_log import log as logging


CONF = cfg.CONF
LOG = logging.getLogger(__name__)

_opts = [
    cfg.ListOpt('enabled_devices',
                default=[],
                sample_default=['common-example',
                                'netconf-openconfig-example'],
                help=('Enabled devices for which the plugin should manage'
                      'configuration. Driver specific configuration for each '
                      'device must be added in separate sections.')),
]

_device_opts = [
    cfg.StrOpt('driver',
               help='The driver to use when configuring the device'),
    cfg.StrOpt('switch_id',
               help='The switch ID, MAC address of the device.'),
    cfg.StrOpt('switch_info',
               help=('Optional string field to be used to store any '
                     'vendor-specific information.')),
    cfg.ListOpt('physical_networks',
                default=[],
                help='A list of physical networks mapped to this device.'),
    cfg.BoolOpt('manage_vlans',
                default=True,
                help=('Set this to False for the device if VLANs should not '
                      'be create and deleted on the device.')),
    ]

networking_baremetal_group = cfg.OptGroup(
    name='networking_baremetal',
    title='ML2 networking-baremetal options')
CONF.register_group(networking_baremetal_group)
CONF.register_opts(_opts, group=networking_baremetal_group)
for device in CONF.networking_baremetal.enabled_devices:
    group = cfg.OptGroup(
        name=device,
        title=f'{device} Device options')
    CONF.register_group(group)
    CONF.register_opts(_device_opts, group=group)


def list_opts():
    return [('networking_baremetal', _opts)]


def list_common_device_driver_opts():
    return [('networking_baremetal', _opts),
            ('common-example', _device_opts)]


def get_devices():
    """Get enabled network devices from configuration

    This is called during driver initialization, during initialization
    additional driver specific configuration is loaded and the drivers
    validation method is called.
    """
    devices = dict()
    for dev in CONF.networking_baremetal.enabled_devices:
        if not CONF[dev].driver:
            LOG.error('IGNORING invalid device %s, driver not specified.', dev)
        if not CONF[dev].switch_id and not CONF[dev].switch_info:
            LOG.error('IGNORING invalid device %s, switch_id and/or '
                      'switch_info is required', dev)

        if CONF[dev].switch_id:
            devices[CONF[dev].switch_id] = dev
        if CONF[dev].switch_info:
            devices[CONF[dev].switch_info] = dev

    return devices



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock
from xml.etree import ElementTree

from oslotest import base

from networking_baremetal import constants
from networking_baremetal.openconfig.lacp import lacp


class TestOpenConfigLACP(base.BaseTestCase):

    @mock.patch.object(lacp, 'LACPInterfaces', autospec=True)
    def test_openconfig_lacp(self, mock_lcap_ifaces):
        mock_lcap_ifaces.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-lacp-interfaces'))
        oc_lacp = lacp.LACP()
        mock_lcap_ifaces.assert_called_with()
        mock_lcap_ifaces.return_value.__len__.return_value = 1
        element = oc_lacp.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = (f'<lacp xmlns="{oc_lacp.NAMESPACE}">'
                    '<fake-lacp-interfaces />'
                    '</lacp>')
        self.assertEqual(expected, xml_str)

    @mock.patch.object(lacp, 'LACPInterface', autospec=True)
    def test_openconfig_lacp_interfaces(self, mock_lacp_iface):
        mock_lacp_iface.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-lacp-interface'))
        oc_lacp_ifaces = lacp.LACPInterfaces()
        self.assertEqual([], oc_lacp_ifaces.interfaces)
        oc_lacp_iface = oc_lacp_ifaces.add('lacp-iface-name')
        mock_lacp_iface.assert_called_with('lacp-iface-name')
        self.assertEqual([oc_lacp_iface], oc_lacp_ifaces.interfaces)
        element = oc_lacp_ifaces.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<interfaces>'
                    '<fake-lacp-interface />'
                    '</interfaces>')
        self.assertEqual(expected, xml_str)

    @mock.patch.object(lacp, 'LACPInterfaceConfig', autospec=True)
    def test_openconfig_lacp_interface(self, mock_lacp_if_conf):
        mock_lacp_if_conf.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-lacp-interface-config'))
        self.assertRaises(TypeError, lacp.LACPInterface, int(20))
        oc_lacp_iface = lacp.LACPInterface('lacp-iface-name')
        self.assertEqual('lacp-iface-name', oc_lacp_iface.name)
        element = oc_lacp_iface.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<interface operation="merge">'
                    '<name>lacp-iface-name</name>'
                    '<fake-lacp-interface-config />'
                    '</interface>')
        self.assertEqual(expected, xml_str)
        oc_lacp_iface.operation = constants.NetconfEditConfigOperation.REMOVE
        element = oc_lacp_iface.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<interface operation="remove">'
                    '<name>lacp-iface-name</name>'
                    '<fake-lacp-interface-config />'
                    '</interface>')
        self.assertEqual(expected, xml_str)

    def test_openconfig_lacp_interface_config(self):
        self.assertRaises(ValueError,
                          lacp.LACPInterfaceConfig, 'name',
                          **dict(operation='invalid'))
        self.assertRaises(ValueError, lacp.LACPInterfaceConfig,
                          'name', **dict(interval='invalid'))
        self.assertRaises(ValueError, lacp.LACPInterfaceConfig,
                          'name', **dict(lacp_mode='invalid'))
        lacp_if_conf = lacp.LACPInterfaceConfig('lacp-iface-name')
        element = lacp_if_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config operation="merge">'
                    '<name>lacp-iface-name</name>'
                    '<interval>SLOW</interval>'
                    '<lacp-mode>ACTIVE</lacp-mode>'
                    '</config>')
        self.assertEqual(expected, xml_str)
        lacp_if_conf.interval = constants.LACP_PERIOD_FAST
        lacp_if_conf.lacp_mode = constants.LACP_ACTIVITY_PASSIVE
        element = lacp_if_conf.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<config operation="merge">'
                    '<name>lacp-iface-name</name>'
                    '<interval>FAST</interval>'
                    '<lacp-mode>PASSIVE</lacp-mode>'
                    '</config>')
        self.assertEqual(expected, xml_str)



# Copyright 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.db import provisioning_blocks
from neutron.plugins.ml2.drivers import mech_agent
from neutron_lib.api.definitions import portbindings
from neutron_lib.api.definitions import provider_net
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib.plugins.ml2 import api
from oslo_config import cfg
from oslo_log import log as logging

from networking_baremetal import common
from networking_baremetal import config
from networking_baremetal import constants
from networking_baremetal import exceptions

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

BAREMETAL_DRV_ENTITY = 'BAREMETAL_DRV_ENTITIY'


class BaremetalMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):

    def __init__(self):
        super(BaremetalMechanismDriver, self).__init__(
            agent_type=constants.BAREMETAL_AGENT_TYPE,
            vif_type=portbindings.VIF_TYPE_OTHER,
            vif_details={
                portbindings.VIF_DETAILS_CONNECTIVITY: self.connectivity},
            supported_vnic_types=[portbindings.VNIC_BAREMETAL])

        self.devices = config.get_devices()
        # Use set to remove duplicates,
        # i.e device has both switch_id and switch_info
        for device_id in set(self.devices.values()):
            device_driver = common.driver_mgr(device_id)
            device_driver.load_config()
            try:
                device_driver.validate()
            except exceptions.DriverValidationError as e:
                LOG.exception(e)

    @property
    def connectivity(self):
        return portbindings.CONNECTIVITY_L2

    def get_allowed_network_types(self, agent):
        """Return the agent's or driver's allowed network types.

        For example: return ('flat', ...). You can also refer to the
        configuration the given agent exposes.
        """
        return [n_const.TYPE_FLAT, n_const.TYPE_VLAN]

    def get_mappings(self, agent):
        """Return the agent's bridge or interface mappings.

        For example: agent['configurations'].get('bridge_mappings', {}).
        """
        return agent['configurations'].get('bridge_mappings', {})

    def create_network_precommit(self, context):
        """Allocate resources for a new network.

        Create a new network, allocating resources as necessary in the
        database. Called inside transaction context on session. Call
        cannot block.  Raising an exception will result in a rollback
        of the current transaction.

        :param context: NetworkContext instance describing the new
            network.
        """
        pass

    def create_network_postcommit(self, context):
        """Create a network.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        :param context: NetworkContext instance describing the new
            network.
        """
        network = context.current
        network_type = network[provider_net.NETWORK_TYPE]
        segmentation_id = network[provider_net.SEGMENTATION_ID]
        physical_network = network[provider_net.PHYSICAL_NETWORK]

        # If not VLAN network, or no segmentation_id - nothing to do.
        if network_type != n_const.TYPE_VLAN or not segmentation_id:
            return

        # TODO(hjensas): This should be parallelized
        for device in CONF.networking_baremetal.enabled_devices:
            # VLAN management is disabled for this device
            if not CONF[device].manage_vlans:
                continue
            # Skip device if not on physical network
            if not self._is_device_on_physnet(device, physical_network):
                continue

            driver = common.driver_mgr(device)
            driver.create_network(context)

    def update_network_precommit(self, context):
        """Update resources of a network.

        Update values of a network, updating the associated resources
        in the database. Called inside transaction context on session.
        Raising an exception will result in rollback of the
        transaction.
        update_network_precommit is called for all changes to the
        network state. It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.

        :param context: NetworkContext instance describing the new
            state of the network, as well as the original state prior
            to the update_network call.
        """
        pass

    def update_network_postcommit(self, context):
        """Update a network.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.
        update_network_postcommit is called for all changes to the
        network state.  It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.

        :param context: NetworkContext instance describing the new
            state of the network, as well as the original state prior
            to the update_network call.
        """
        network = context.current
        network_orig = context.original
        network_type = network[provider_net.NETWORK_TYPE]
        segmentation_id = network[provider_net.SEGMENTATION_ID]
        network_type_orig = network_orig[provider_net.NETWORK_TYPE]
        physical_network = network[provider_net.PHYSICAL_NETWORK]

        if (network_type != n_const.TYPE_VLAN
                and network_type_orig != n_const.TYPE_VLAN):
            return

        if not segmentation_id and not network_type_orig:
            return

        # TODO(hjensas): This should be parallelized
        for device in CONF.networking_baremetal.enabled_devices:
            # VLAN management is disabled for this device
            if not CONF[device].manage_vlans:
                continue
            # Skip device if not on physical network
            if not self._is_device_on_physnet(device, physical_network):
                continue

            driver = common.driver_mgr(device)
            driver.update_network(context)

    def delete_network_precommit(self, context):
        """Delete resources for a network.

        Delete network resources previously allocated by this
        mechanism driver for a network. Called inside transaction
        context on session. Runtime errors are not expected, but
        raising an exception will result in rollback of the
        transaction.

        :param context: NetworkContext instance describing the current
            state of the network, prior to the call to delete it.
        """
        pass

    def delete_network_postcommit(self, context):
        """Delete a network.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.

        :param context: NetworkContext instance describing the current
            state of the network, prior to the call to delete it.
        """
        network = context.current
        network_type = network[provider_net.NETWORK_TYPE]
        segmentation_id = network[provider_net.SEGMENTATION_ID]
        physical_network = network[provider_net.PHYSICAL_NETWORK]

        # If not VLAN network, or no segmentation_id - nothing to do.
        if network_type != n_const.TYPE_VLAN or not segmentation_id:
            return

        # TODO(hjensas): This should be parallelized
        for device in CONF.networking_baremetal.enabled_devices:
            # VLAN management is disabled for this device
            if not CONF[device].manage_vlans:
                continue
            # Skip device if not on physical network
            if not self._is_device_on_physnet(device, physical_network):
                continue

            driver = common.driver_mgr(device)
            driver.delete_network(context)

    def create_subnet_precommit(self, context):
        """Allocate resources for a new subnet.

        Create a new subnet, allocating resources as necessary in the
        database. Called inside transaction context on session. Call
        cannot block.  Raising an exception will result in a rollback
        of the current transaction.

        :param context: SubnetContext instance describing the new
            subnet.
        """
        pass

    def create_subnet_postcommit(self, context):
        """Create a subnet.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.

        :param context: SubnetContext instance describing the new
            subnet.
        """
        pass

    def update_subnet_precommit(self, context):
        """Update resources of a subnet.

        Update values of a subnet, updating the associated resources
        in the database. Called inside transaction context on session.
        Raising an exception will result in rollback of the
        transaction.
        update_subnet_precommit is called for all changes to the
        subnet state. It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.

        :param context: SubnetContext instance describing the new
            state of the subnet, as well as the original state prior
            to the update_subnet call.
        """
        pass

    def update_subnet_postcommit(self, context):
        """Update a subnet.

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Raising an exception will
        cause the deletion of the resource.
        update_subnet_postcommit is called for all changes to the
        subnet state.  It is up to the mechanism driver to ignore
        state or state changes that it does not know or care about.

        :param context: SubnetContext instance describing the new
            state of the subnet, as well as the original state prior
            to the update_subnet call.
        """
        pass

    def delete_subnet_precommit(self, context):
        """Delete resources for a subnet.

        Delete subnet resources previously allocated by this
        mechanism driver for a subnet. Called inside transaction
        context on session. Runtime errors are not expected, but
        raising an exception will result in rollback of the
        transaction.

        :param context: SubnetContext instance describing the current
            state of the subnet, prior to the call to delete it.
        """
        pass

    def delete_subnet_postcommit(self, context):
        """Delete a subnet.a

        Called after the transaction commits. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance. Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.

        :param context: SubnetContext instance describing the current
            state of the subnet, prior to the call to delete it.
        """
        pass

    def create_port_precommit(self, context):
        """Allocate resources for a new port.

        Create a new port, allocating resources as necessary in the
        database. Called inside transaction context on session. Call
        cannot block.  Raising an exception will result in a rollback
        of the current transaction.

        :param context: PortContext instance describing the port.
        """
        pass

    def create_port_postcommit(self, context):
        """Create a port.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Raising an exception will
        result in the deletion of the resource.

        :param context: PortContext instance describing the port.
        """
        pass

    def update_port_precommit(self, context):
        """Update resources of a port.

        Called inside transaction context on session to complete a
        port update as defined by this mechanism driver. Raising an
        exception will result in rollback of the transaction.
        update_port_precommit is called for all changes to the port
        state. It is up to the mechanism driver to ignore state or
        state changes that it does not know or care about.

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        """
        pass

    def update_port_postcommit(self, context):
        """Update a port.

        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Raising an exception will
        result in the deletion of the resource.
        update_port_postcommit is called for all changes to the port
        state. It is up to the mechanism driver to ignore state or
        state changes that it does not know or care about.

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        """
        port = context.current
        port_orig = context.original
        if self._is_bound(port):
            if port_orig:
                self._update_port(context)
            provisioning_blocks.provisioning_complete(
                context._plugin_context, port['id'], resources.PORT,
                BAREMETAL_DRV_ENTITY)
        elif self._is_bound(port_orig):
            # The port has been unbound. This will cause the local link
            # information to be lost, so remove the port from the network on
            # the switch now while we have the required information.
            self._unplug_port(context, current=False)

    def delete_port_precommit(self, context):
        """Delete resources of a port.

        Called inside transaction context on session. Runtime errors
        are not expected, but raising an exception will result in
        rollback of the transaction.

        :param context: PortContext instance describing the current
            state of the port, prior to the call to delete it.
        """
        pass

    def delete_port_postcommit(self, context):
        """Delete a port.

        state of the port, prior to the call to delete it.
        Called after the transaction completes. Call can block, though
        will block the entire process so care should be taken to not
        drastically affect performance.  Runtime errors are not
        expected, and will not prevent the resource from being
        deleted.

        :param context: PortContext instance describing the current
            state of the port, prior to the call to delete it.
        """
        self._unplug_port(context)

    def try_to_bind_segment_for_agent(self, context, segment, agent):
        """Try to bind with segment for agent.

        :param context: PortContext instance describing the port
        :param segment: segment dictionary describing segment to bind
        :param agent: agents_db entry describing agent to bind
        :returns: True iff segment has been bound for agent

        Neutron segments api-ref:
          https://docs.openstack.org/api-ref/network/v2/#segments

        Example segment dictionary: {'segmentation_id': 'segmentation_id',
                                     'network_type': 'network_type',
                                     'id': 'segment_uuid'}

        Called outside any transaction during bind_port() so that
        derived MechanismDrivers can use agent_db data along with
        built-in knowledge of the corresponding agent's capabilities
        to attempt to bind to the specified network segment for the
        agent.

        If the segment can be bound for the agent, this function must
        call context.set_binding() with appropriate values and then
        return True. Otherwise, it must return False.
        """
        if not self.check_segment_for_agent(segment, agent):
            return False

        port = context.current
        binding_profile = port[portbindings.PROFILE] or {}
        local_link_information = binding_profile.get(
            constants.LOCAL_LINK_INFO, [])
        local_group_information = binding_profile.get(
            constants.LOCAL_GROUP_INFO, {})
        bond_mode = local_group_information.get('bond_mode')

        by_device = {}
        for link in local_link_information:
            device = self._get_device(link)

            # If there is no device for all links the port will be bound, this
            # keeps backward compatibility.
            if not device:
                continue

            # Device was found, but no port_id in link - fail port binding
            if not link.get(constants.PORT_ID):
                LOG.warning('Cannot bind port %(port)s. no port_id in link '
                            'information: %(link)s', {'port': port[api.ID],
                                                      'link': link})
                return False

            # Check device on physnet, if not fail port binding
            if not self._is_device_on_physnet(device,
                                              segment[api.PHYSICAL_NETWORK]):
                LOG.warning(
                    'Cannot bind port %(port)s, device %(device)s is '
                    'not on physical network %(physnet)s',
                    {'port': port[api.ID], 'device': device,
                     'physnet': segment[api.PHYSICAL_NETWORK]})
                return False

            by_device.setdefault(device, {})
            by_device[device].setdefault('links', [])

            if 'driver' not in by_device[device]:
                # Load the driver, fail port binding on load error
                try:
                    driver = common.driver_mgr(device)
                    by_device[device]['driver'] = driver
                except exceptions.DriverEntrypointLoadError as e:
                    LOG.warning('Cannot bind port %(port)s, failed to load '
                                'driver for device %(device)s',
                                {'link': link, 'port': port[api.ID],
                                 'device': device})
                    LOG.debug(e.message)
                    return False

            by_device[device]['links'].append(link)

        if not by_device:
            # NOTE(vsaienko): we can call set_binding ONLY when we complete
            # binding for the port in the segment. We do not handle the port
            # and want to let other drivers to bind it.
            return False

        # Check driver(s) support the bond_mode - if not fail port binding
        if (bond_mode and by_device
                and not self._is_bond_mode_supported(bond_mode, by_device)):
            LOG.warning('Cannot bind port %(port)s, unsupported '
                        'bond_mode %(bond_mode)s',
                        {'port': port[api.ID], 'bond_mode': bond_mode})
            return False

        # Call each drivers create_port method to plug the device links
        for device, args in by_device.items():
            driver = args['driver']
            driver.create_port(context, segment, args['links'])

        # Complete the port binding
        provisioning_blocks.add_provisioning_component(
            context._plugin_context, port[api.ID], resources.PORT,
            BAREMETAL_DRV_ENTITY)
        context.set_binding(segment[api.ID],
                            self.get_vif_type(context, agent, segment),
                            self.get_vif_details(context, agent, segment))

        return True

    def _is_bound(self, context):
        """Check if port is currently bound by this driver

        :param context: Port context
        :returns: True/False
        """
        return (context[portbindings.VNIC_TYPE] in self.supported_vnic_types
                and context[portbindings.VIF_TYPE] == self.vif_type)

    @staticmethod
    def _is_device_on_physnet(device, physical_network):
        """Check if Device is connected to physical network

        If the device is not configured to any physical networks, return
        True so that all networks are created on the switch.

        :param device: Netconf device in config
        :param physical_network: Physical network
        :returns: True or False
        """
        if (CONF[device].physical_networks
                and physical_network not in CONF[device].physical_networks):
            return False

        return True

    def _get_device(self, link):
        """Get device identifier from link information

        :param link: Link information
        :returns: Device identifier (switch_id or switch_info)
        """
        device = None
        switch_id = link.get(constants.SWITCH_ID)
        switch_info = link.get(constants.SWITCH_INFO)
        if switch_id and switch_id in self.devices:
            device = self.devices[switch_id]
        elif switch_info and switch_info in self.devices:
            device = self.devices[switch_info]

        return device

    def _update_port(self, context):
        """Update port

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        """
        port = context.current
        binding_profile = port[portbindings.PROFILE]
        local_link_information = binding_profile.get(
            constants.LOCAL_LINK_INFO, [])
        local_group_information = binding_profile.get(
            constants.LOCAL_GROUP_INFO, {})
        bond_mode = local_group_information.get('bond_mode')
        by_device = {}
        for link in local_link_information:
            device = self._get_device(link)
            # No device == noop
            if not device:
                continue

            by_device.setdefault(device, {})
            by_device[device].setdefault('links', [])

            if 'driver' not in by_device[device]:
                # Load the driver, if this fails the link cannot be updated
                try:
                    driver = common.driver_mgr(device)
                    by_device[device]['driver'] = driver
                except exceptions.DriverEntrypointLoadError as e:
                    LOG.warning('Cannot update link %(link)s on port '
                                '%(port)s, failed to load driver for device '
                                '%(device)s', {'link': link,
                                               'port': port[api.ID],
                                               'device': device})
                    LOG.debug(e.message)
                    continue

            by_device[device]['links'].append(link)

        # Check driver(s) support the bond_mode
        if (bond_mode and by_device
                and not self._is_bond_mode_supported(bond_mode, by_device)):
            LOG.error('Cannot update port %(port)s on device, unsupported '
                      'bond_mode %(bond_mode)s', {'port': port[api.ID],
                                                  'bond_mode': bond_mode})
            return

        # Call each drivers update_port method
        for device, args in by_device.items():
            driver = args['driver']
            driver.update_port(context, args['links'])

    def _unplug_port(self, context, current=True):
        """Unplug/Unbind/Delete port

        :param context: PortContext instance describing the new
            state of the port, as well as the original state prior
            to the update_port call.
        :param current: Boolean, when true use context.current, when
            false use context.original
        """
        if current:
            port = context.current
        else:
            port = context.original

        binding_profile = port[portbindings.PROFILE] or {}
        local_link_information = binding_profile.get(
            constants.LOCAL_LINK_INFO, [])
        local_group_information = binding_profile.get(
            constants.LOCAL_GROUP_INFO, {})
        bond_mode = local_group_information.get('bond_mode')

        by_device = {}
        for link in local_link_information:
            device = self._get_device(link)
            # No device == noop
            if not device:
                continue

            by_device.setdefault(device, {})
            by_device[device].setdefault('links', [])

            if 'driver' not in by_device[device]:
                # Load the driver, if this fails the link cannot be unbound
                try:
                    driver = common.driver_mgr(device)
                    by_device[device]['driver'] = driver
                except exceptions.DriverEntrypointLoadError as e:
                    LOG.warning('Cannot delete link %(link)s for port '
                                '%(port)s, failed to load driver for device '
                                '%(device)s', {'link': link,
                                               'port': port[api.ID],
                                               'device': device})
                    LOG.debug(e.message)
                    continue

            by_device[device]['links'].append(link)

        # Check driver(s) support the bond_mode
        if (bond_mode and by_device
                and not self._is_bond_mode_supported(bond_mode, by_device)):
            LOG.warning('Cannot delete port %(port)s on device, unsupported '
                        'bond_mode %(bond_mode)s', {'port': port[api.ID],
                                                    'bond_mode': bond_mode})
            return

        # Call each drivers delete_port method to unplug the device links
        for device, args in by_device.items():
            driver = args['driver']
            driver.delete_port(context, args['links'], current=current)

    @staticmethod
    def _is_bond_mode_supported(bond_mode, by_device):
        """Check if drivers support the bond mode

        :param bond_mode: The bond mode
        :param by_device: Dictionary of driver and links per-device.
        """
        for device, args in by_device.items():
            driver = args['driver']

            if bond_mode and bond_mode not in driver.SUPPORTED_BOND_MODES:
                return False

        return True



# Copyright (c) 2017 Mirantis, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


from oslo_utils import uuidutils


def get_test_network(**kw):
    """Return a network object with appropriate attributes."""

    result = {
        "provider:physical_network": kw.get("physical_network", "mynetwork"),
        "ipv6_address_scope": kw.get("ipv6_address_scope", None),
        "revision_number": kw.get("revision_number", 7),
        "port_security_enabled": kw.get("port_security_enabled", True),
        "mtu": kw.get("mtu", 1500),
        "id": kw.get("id", uuidutils.generate_uuid()),
        "router:external": kw.get("router:external", False),
        "availability_zone_hints": kw.get("availability_zone_hints", []),
        "availability_zones": kw.get("availability_zones", ["nova"]),
        "ipv4_address_scope": kw.get("ipv4_address_scope", None),
        "shared": kw.get("shared", False),
        "project_id": kw.get("project_id", uuidutils.generate_uuid()),
        "status": kw.get("status", "ACTIVE"),
        "subnets": kw.get("subnets", []),
        "description": kw.get("description", ""),
        "tags": kw.get("tags", []),
        "provider:segmentation_id": kw.get("segmentation_id", 113),
        "name": kw.get("name", "private"),
        "admin_state_up": kw.get("admin_state_up", True),
        "tenant_id": kw.get("tenant_id", uuidutils.generate_uuid()),
        "provider:network_type": kw.get("network_type", "flat")
    }
    return result


def get_test_port(network_id, **kw):
    """Return a port object with appropriate attributes."""

    result = {
        "status": kw.get("status", "DOWN"),
        "binding:host_id": kw.get("host_id", "aaa.host"),
        "description": kw.get("description", ""),
        "allowed_address_pairs": kw.get("allowed_address_pairs", []),
        "tags": kw.get("tags", []),
        "extra_dhcp_opts": kw.get("extra_dhcp_opts", []),
        "device_owner": kw.get("device_owner", "baremetal:host"),
        "revision_number": kw.get("revision_number", 7),
        "port_security_enabled": kw.get("port_security_enabled", False),
        "binding:profile": kw.get("binding_profile",
                                  {'local_link_information': [
                                      {'switch_info': 'foo',
                                       'port_id': 'Gig0/1',
                                       'switch_id': 'aa:bb:cc:dd:ee:ff'}]}),
        "fixed_ips": kw.get("fixed_ips", []),
        "id": kw.get("id", uuidutils.generate_uuid()),
        "security_groups": kw.get("security_groups", []),
        "device_id": kw.get("device_id", ""),
        "name": kw.get("name", "Port1"),
        "admin_state_up": kw.get("admin_state_up", True),
        "network_id": network_id,
        "tenant_id": kw.get("tenant_id", uuidutils.generate_uuid()),
        "binding:vif_details": kw.get("vif_details", {}),
        "binding:vnic_type": kw.get("vnic_type", "baremetal"),
        "binding:vif_type": kw.get("vif_type", "unbound"),
        "mac_address": kw.get("mac_address", "fa:16:3e:c2:2a:8f"),
        "project_id": kw.get("project_id", uuidutils.generate_uuid())
    }
    return result


def get_test_subnet(network_id, **kw):
    """Return a subnet object with appropriate attributes."""

    result = {
        "service_types": kw.get("service_types", []),
        "description": kw.get("description", ""),
        "enable_dhcp": kw.get("enable_dhcp", True),
        "tags": kw.get("tags", []),
        "network_id": network_id,
        "tenant_id": kw.get("tenant_id", uuidutils.generate_uuid()),
        "dns_nameservers": kw.get("dns_nameservers", []),
        "gateway_ip": kw.get("gateway_ip", "10.1.0.1"),
        "ipv6_ra_mode": kw.get("ipv6_ra_mode", None),
        "allocation_pools": kw.get("allocation_pools", [{"start": "10.1.0.2",
                                                         "end": "10.1.0.62"}]),
        "host_routes": kw.get("host_routes", []),
        "revision_number": kw.get("revision_number", 7),
        "ip_version": kw.get("ip_version", 4),
        "ipv6_address_mode": kw.get("ipv6_address_mode", None),
        "cidr": kw.get("cidr", "10.1.0.0/26"),
        "project_id": kw.get("project_id", uuidutils.generate_uuid()),
        "id": kw.get("id", uuidutils.generate_uuid()),
        "subnetpool_id": kw.get("subnetpool_id", uuidutils.generate_uuid()),
        "name": kw.get("name", "subnet0")
    }
    return result


def get_test_segment(**kw):
    result = {
        'segmentation_id': kw.get('segmentation_id', '123'),
        'network_type': kw.get('network_type', 'flat'),
        'id': uuidutils.generate_uuid()
    }
    return result



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock
from xml.etree import ElementTree

from oslotest import base

from networking_baremetal.openconfig.network_instance import network_instance
from networking_baremetal.openconfig.vlan import vlan


class TestNetworkInstance(base.BaseTestCase):

    @mock.patch.object(network_instance, 'NetworkInstance', autospec=True)
    def test_network_instances(self, mock_net_instance):
        mock_net_instance.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-net-instance'))
        net_instances = network_instance.NetworkInstances()
        net_instance = net_instances.add('default')
        self.assertEqual([net_instance], net_instances.network_instances)
        element = net_instances.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = (f'<network-instances xmlns="{net_instances.NAMESPACE}">'
                    '<fake-net-instance />'
                    '</network-instances>')
        self.assertEqual(expected, xml_str)

    @mock.patch.object(vlan, 'Vlans', autospec=True)
    def test_network_instance(self, mock_oc_vlans):
        mock_oc_vlans.return_value.to_xml_element.return_value = (
            ElementTree.Element('fake-oc-vlans'))
        mock_oc_vlans.return_value.__len__.return_value = 1
        net_instance = network_instance.NetworkInstance('default')
        self.assertEqual(mock_oc_vlans(), net_instance.vlans)
        element = net_instance.to_xml_element()
        xml_str = ElementTree.tostring(element).decode("utf-8")
        expected = ('<network-instance>'
                    '<name>default</name>'
                    '<fake-oc-vlans />'
                    '</network-instance>')
        self.assertEqual(expected, xml_str)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from unittest import mock
from xml.etree import ElementTree

from ncclient import manager
from neutron.plugins.ml2 import driver_context
from neutron_lib import constants as n_const
from neutron_lib.plugins.ml2 import api
from oslo_config import fixture as config_fixture
from oslo_utils import uuidutils

from networking_baremetal import config
from networking_baremetal import constants
from networking_baremetal.constants import NetconfEditConfigOperation as nc_op
from networking_baremetal.drivers.netconf import openconfig
from networking_baremetal.openconfig.interfaces import interfaces
from networking_baremetal.openconfig.lacp import lacp
from networking_baremetal.tests import base
from networking_baremetal.tests.unit.plugins.ml2 import utils as ml2_utils


OC_IF_NS = 'http://openconfig.net/yang/interfaces'
OC_IF_ETH_NS = 'http://openconfig.net/yang/interfaces/ethernet'
OC_IF_AGG_NS = 'http://openconfig.net/yang/interfaces/aggregate'
XML_IFACES_AGGREDATE_ID = f'''
<data>
  <interfaces xmlns="{OC_IF_NS}">
    <interface>
      <name>foo1/1</name>
      <ethernet xmlns="{OC_IF_ETH_NS}">
        <config>
          <aggregate-id xmlns="{OC_IF_AGG_NS}">Po10</aggregate-id>
        </config>
      </ethernet>
    </interface>
    <interface>
      <name>foo1/2</name>
      <ethernet xmlns="{OC_IF_ETH_NS}">
        <config>
          <aggregate-id xmlns="{OC_IF_AGG_NS}">Po10</aggregate-id>
        </config>
      </ethernet>
    </interface>
  </interfaces>
</data>
'''

XML_AGGREGATE_IFACES = f'''
<data>
  <interfaces xmlns="{OC_IF_NS}">
    <interface>
      <name>Po5</name>
    </interface>
    <interface>
      <name>Po7</name>
    </interface>
    <interface>
      <name>Po9</name>
    </interface>
    <interface>
      <name>foo1/1</name>
    </interface>
  </interfaces>
</data>
'''


class TestNetconfOpenConfigClient(base.TestCase):

    def setUp(self):
        super(TestNetconfOpenConfigClient, self).setUp()
        self.device = 'foo'
        self.conf = self.useFixture(config_fixture.Config())
        self.conf.register_opts(config._opts + config._device_opts,
                                group='foo')
        self.conf.register_opts((openconfig._DEVICE_OPTS
                                 + openconfig._NCCLIENT_OPTS), group='foo')
        self.conf.config(enabled_devices=['foo'],
                         group='networking_baremetal')
        self.conf.config(driver='test-driver',
                         switch_id='aa:bb:cc:dd:ee:ff',
                         switch_info='foo',
                         physical_networks=['fake_physical_network'],
                         device_params={'name': 'default'},
                         host='foo.example.com',
                         key_filename='/test/test_key_file',
                         username='foo_user',
                         group='foo')

        self.client = openconfig.NetconfOpenConfigClient(self.device)

    def test_get_lock_session_id(self):
        err_info = (
            '<?xml version="1.0" encoding="UTF-8"?>'
            '<error-info xmlns="urn:ietf:params:xml:ns:netconf:base:1.0">'
            '<session-id>{}</session-id>'
            '</error-info>')
        self.assertEqual('0', self.client._get_lock_session_id(
            err_info.format(0)))
        self.assertEqual('abc-123', self.client._get_lock_session_id(
            err_info.format('abc-123')))

    def test_get_client_args(self):
        self.assertEqual(
            {'device_params': {'name': 'default'},
             'host': 'foo.example.com',
             'hostkey_verify': True,
             'keepalive': True,
             'key_filename': '/test/test_key_file',
             'port': 830,
             'username': 'foo_user',
             'allow_agent': True,
             'look_for_keys': True}, self.client.get_client_args())

    @mock.patch.object(manager, 'connect', autospec=True)
    def test_get_capabilities(self, mock_manager):
        fake_caps = set(constants.IANA_NETCONF_CAPABILITIES.values())
        fake_caps.add('http://openconfig.net/yang/'
                      'network-instance?'
                      'module=openconfig-network-instance&'
                      'revision=2021-07-22')
        fake_caps.add('http://openconfig.net/yang/'
                      'interfaces?'
                      'module=openconfig-interfaces&'
                      'revision=2021-04-06')
        mock_ncclient = mock.Mock()
        mock_ncclient.server_capabilities = fake_caps
        mock_manager.return_value.__enter__.return_value = mock_ncclient
        self.assertEqual({
            ':base:1.0', ':base:1.1', ':candidate', ':confirmed-commit',
            ':confirmed-commit:1.1', ':rollback-on-error', ':startup',
            ':validate', ':validate:1.1', ':writable-running',
            'openconfig-network-instance', 'openconfig-interfaces'},
            self.client.get_capabilities())

    @mock.patch.object(manager, 'connect', autospec=True)
    @mock.patch.object(openconfig.NetconfOpenConfigClient,
                       'get_lock_and_configure', autospec=True)
    def test_edit_config_writable_running(self, mock_lock_config,
                                          mock_manager):
        fake_config = mock.Mock()
        fake_config.to_xml_element.return_value = ElementTree.Element('fake')
        mock_ncclient = mock.Mock()
        fake_caps = {constants.IANA_NETCONF_CAPABILITIES[':writable-running']}
        mock_ncclient.server_capabilities = fake_caps
        mock_manager.return_value.__enter__.return_value = mock_ncclient
        self.client.edit_config(fake_config)
        mock_lock_config.assert_called_once_with(self.client, mock_ncclient,
                                                 openconfig.RUNNING,
                                                 [fake_config], False)

    @mock.patch.object(manager, 'connect', autospec=True)
    @mock.patch.object(openconfig.NetconfOpenConfigClient,
                       'get_lock_and_configure', autospec=True)
    def test_edit_config_candidate(self, mock_lock_config, mock_manager):
        fake_config = mock.Mock()
        fake_config.to_xml_element.return_value = ElementTree.Element('fake')
        mock_ncclient = mock.Mock()
        fake_caps = {constants.IANA_NETCONF_CAPABILITIES[':candidate']}
        mock_ncclient.server_capabilities = fake_caps
        mock_manager.return_value.__enter__.return_value = mock_ncclient
        self.client.edit_config(fake_config)
        mock_lock_config.assert_called_once_with(self.client, mock_ncclient,
                                                 openconfig.CANDIDATE,
                                                 [fake_config], False)

    def test_get_lock_and_configure_confirmed_commit(self):
        self.client.capabilities = {':candidate', ':writable-running',
                                    ':confirmed-commit'}
        fake_config = mock.Mock()
        fake_config.to_xml_element.return_value = ElementTree.Element('fake')
        mock_client = mock.MagicMock()
        self.client.get_lock_and_configure(mock_client, openconfig.CANDIDATE,
                                           [fake_config], False)
        mock_client.locked.assert_called_with(openconfig.CANDIDATE)
        mock_client.discard_changes.assert_called_once()
        mock_client.edit_config.assert_called_with(
            target=openconfig.CANDIDATE,
            config='<config><fake /></config>')
        mock_client.validate.assert_not_called()
        mock_client.commit.assert_has_calls([
            mock.call(confirmed=True, timeout=str(30)), mock.call()])

    def test_get_lock_and_configure_validate(self):
        self.client.capabilities = {':candidate', ':writable-running',
                                    ':validate'}
        fake_config = mock.Mock()
        fake_config.to_xml_element.return_value = ElementTree.Element('fake')
        mock_client = mock.MagicMock()
        self.client.get_lock_and_configure(mock_client, openconfig.CANDIDATE,
                                           [fake_config], False)
        mock_client.locked.assert_called_with(openconfig.CANDIDATE)
        mock_client.discard_changes.assert_called_once()
        mock_client.edit_config.assert_called_with(
            target=openconfig.CANDIDATE,
            config='<config><fake /></config>')
        mock_client.validate.assert_called_once_with(
            source=openconfig.CANDIDATE)
        mock_client.commit.assert_called_once_with()

    def test_get_lock_and_configure_writeable_running(self):
        self.client.capabilities = {':writable-running'}
        fake_config = mock.Mock()
        fake_config.to_xml_element.return_value = ElementTree.Element('fake')
        mock_client = mock.MagicMock()
        self.client.get_lock_and_configure(mock_client, openconfig.RUNNING,
                                           [fake_config], False)
        mock_client.locked.assert_called_with(openconfig.RUNNING)
        mock_client.discard_changes.assert_not_called()
        mock_client.validate.assert_not_called()
        mock_client.commit.assert_not_called()
        mock_client.edit_config.assert_called_with(
            target=openconfig.RUNNING,
            config='<config><fake /></config>')

    @mock.patch.object(manager, 'connect', autospec=True)
    def test_get(self, mock_manager):
        fake_query = interfaces.Interfaces()
        fake_query.add('foo1/1')
        mock_ncclient = mock.Mock()
        mock_manager.return_value.__enter__.return_value = mock_ncclient
        self.client.get(query=fake_query)
        mock_ncclient.get.assert_called_with(filter=('subtree', mock.ANY))

    def test_get_aggregation_ids(self):
        self.conf.config(link_aggregate_prefix='foo',
                         link_aggregate_range='5..10',
                         group=self.device)
        self.assertEqual({'foo5', 'foo6', 'foo7', 'foo8', 'foo9', 'foo10'},
                         self.client.get_aggregation_ids())

    def test_allocate_deferred(self):
        aggregate_id = 'foo5'
        _config = []
        ifaces = interfaces.Interfaces()
        iface_a = ifaces.add('foo1/1',
                             interface_type=constants.IFACE_TYPE_ETHERNET)
        iface_a.ethernet.config.aggregate_id = openconfig.DEFERRED
        iface_b = ifaces.add('foo1/2',
                             interface_type=constants.IFACE_TYPE_ETHERNET)
        iface_b.ethernet.config.aggregate_id = openconfig.DEFERRED
        iface_agg = ifaces.add(openconfig.DEFERRED,
                               interface_type=constants.IFACE_TYPE_AGGREGATE)
        iface_agg.config.name = openconfig.DEFERRED
        _config.append(ifaces)
        _lacp = lacp.LACP()
        lacp_ifaces = lacp.LACPInterfaces()
        lacp_ifaces.add(openconfig.DEFERRED)
        _config.append(_lacp)
        self.client.allocate_deferred(aggregate_id, _config)
        for conf in _config:
            if isinstance(conf, interfaces.Interfaces):
                for iface in ifaces:
                    if isinstance(iface, interfaces.InterfaceAggregate):
                        self.assertEqual(iface.name, aggregate_id)
                        self.assertEqual(iface.config.name, aggregate_id)
                    elif isinstance(iface, interfaces.InterfaceEthernet):
                        self.assertEqual(iface.ethernet.config.aggregate_id,
                                         aggregate_id)
            if isinstance(conf, lacp.LACP):
                for lacp_iface in _lacp.interfaces.interfaces:
                    self.assertEqual(lacp_iface.name, aggregate_id)
                    self.assertEqual(lacp_iface.config.name, aggregate_id)

    def test_get_free_aggregate_id(self):
        self.conf.config(link_aggregate_prefix='Po',
                         link_aggregate_range='5..10',
                         group=self.device)
        mock_get_result = mock.Mock()
        mock_get_result.data_xml = XML_AGGREGATE_IFACES
        mock_client_locked = mock.Mock()
        mock_client_locked.get.return_value = mock_get_result
        used_aggregate_ids = {'Po5', 'Po7', 'Po9'}
        all_aggregate_ids = self.client.get_aggregation_ids()
        result = self.client.get_free_aggregate_id(mock_client_locked)
        self.assertNotIn(result, used_aggregate_ids)
        self.assertIn(result, all_aggregate_ids)


class TestNetconfOpenConfigDriver(base.TestCase):

    def setUp(self):
        super(TestNetconfOpenConfigDriver, self).setUp()
        self.device = 'foo'
        self.conf = self.useFixture(config_fixture.Config())
        self.conf.register_opts(config._opts + config._device_opts,
                                group='foo')
        self.conf.register_opts((openconfig._DEVICE_OPTS
                                 + openconfig._NCCLIENT_OPTS), group='foo')
        self.conf.config(enabled_devices=['foo'],
                         group='networking_baremetal')
        self.conf.config(driver='test-driver',
                         switch_id='aa:bb:cc:dd:ee:ff',
                         switch_info='foo',
                         physical_networks=['fake_physical_network'],
                         device_params={'name': 'default'},
                         host='foo.example.com',
                         key_filename='/test/test_key_file',
                         username='foo_user',
                         group='foo')
        mock_client = mock.patch.object(openconfig, 'NetconfOpenConfigClient',
                                        autospec=True)
        self.mock_client = mock_client.start()
        self.addCleanup(mock_client.stop)

        self.driver = openconfig.NetconfOpenConfigDriver(self.device)
        self.mock_client.assert_called_once_with('foo')
        self.mock_client.reset_mock()

    def test_validate(self):
        self.driver.validate()
        self.driver.client.get_capabilities.assert_called_once_with()

    @mock.patch.object(openconfig, 'CONF', autospec=True)
    def test_load_config(self, mock_conf):
        self.driver.load_config()
        mock_conf.register_opts.assert_has_calls(
            [mock.call(openconfig._DEVICE_OPTS, group=self.driver.device),
             mock.call(openconfig._NCCLIENT_OPTS, group=self.driver.device)])

    def test_create_network(self):
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_nc.current = ml2_utils.get_test_network()
        self.driver.create_network(m_nc)
        net_instances = self.driver.client.edit_config.call_args[0][0]
        for net_instance in net_instances:
            self.assertEqual(net_instance.name, 'default')
            vlans = net_instance.vlans
            for vlan in vlans:
                self.assertEqual(vlan.config.operation, nc_op.MERGE.value)
                self.assertEqual(vlan.config.name,
                                 self.driver._uuid_as_hex(m_nc.current['id']))
                self.assertEqual(vlan.config.status, constants.VLAN_ACTIVE)

    def test_update_network_no_changes(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN)
        m_nc.original = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN)
        self.assertEqual(m_nc.current, m_nc.original)
        self.driver.update_network(m_nc)
        self.driver.client.edit_config.assert_not_called()

    def test_update_network_change_vlan_id(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=10)
        m_nc.original = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=20)
        self.driver.update_network(m_nc)
        call_args_list = self.driver.client.edit_config.call_args_list
        del_net_instances = call_args_list[0][0][0]
        add_net_instances = call_args_list[1][0][0]
        self.driver.client.edit_config.assert_has_calls(
            [mock.call(del_net_instances), mock.call(add_net_instances)])
        for net_instance in del_net_instances:
            self.assertEqual(net_instance.name, 'default')
            for vlan in net_instance.vlans:
                self.assertEqual(vlan.operation, nc_op.REMOVE.value)
                self.assertEqual(vlan.vlan_id, 20)
                self.assertEqual(vlan.config.status, constants.VLAN_SUSPENDED)
                self.assertEqual(vlan.config.name, 'neutron-DELETED-20')
        for net_instance in add_net_instances:
            self.assertEqual(net_instance.name, 'default')
            for vlan in net_instance.vlans:
                self.assertEqual(vlan.operation, nc_op.MERGE.value)
                self.assertEqual(vlan.config.name,
                                 self.driver._uuid_as_hex(network_id))
                self.assertEqual(vlan.vlan_id, 10)

    def test_update_network_change_admin_state(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=10,
            admin_state_up=False)
        m_nc.original = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=10,
            admin_state_up=True)
        self.driver.update_network(m_nc)
        call_args_list = self.driver.client.edit_config.call_args_list
        add_net_instances = call_args_list[0][0][0]
        self.driver.client.edit_config.assert_called_once_with(
            add_net_instances)
        for net_instance in add_net_instances:
            self.assertEqual(net_instance.name, 'default')
            for vlan in net_instance.vlans:
                self.assertEqual(vlan.operation, nc_op.MERGE.value)
                self.assertEqual(vlan.config.status, constants.VLAN_SUSPENDED)
                self.assertEqual(vlan.config.name,
                                 self.driver._uuid_as_hex(network_id))
                self.assertEqual(vlan.vlan_id, 10)

    def test_delete_network(self):
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_nc.current = ml2_utils.get_test_network(
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        self.driver.delete_network(m_nc)
        self.driver.client.edit_config.assert_called_once()
        call_args_list = self.driver.client.edit_config.call_args_list
        net_instances = call_args_list[0][0][0]
        for net_instance in net_instances:
            self.assertEqual(net_instance.name, 'default')
            for vlan in net_instance.vlans:
                self.assertEqual(vlan.operation, nc_op.REMOVE.value)
                self.assertEqual(vlan.vlan_id, 15)
                self.assertEqual(vlan.config.status, constants.VLAN_SUSPENDED)
                self.assertEqual(vlan.config.name, 'neutron-DELETED-15')

    def test_create_port_vlan(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id)
        m_pc.network = m_nc
        segment = {
            api.ID: uuidutils.generate_uuid(),
            api.PHYSICAL_NETWORK:
                m_nc.current['provider:physical_network'],
            api.NETWORK_TYPE: m_nc.current['provider:network_type'],
            api.SEGMENTATION_ID: m_nc.current['provider:segmentation_id']}
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.create_port(m_pc, segment, links)
        self.driver.client.edit_config.assert_called_once()
        call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = call_args_list[0][0][0]
        for iface in ifaces:
            self.assertEqual(iface.name, links[0]['port_id'])
            self.assertEqual(iface.config.enabled,
                             m_pc.current['admin_state_up'])
            self.assertEqual(iface.config.mtu, m_nc.current[api.MTU])
            self.assertEqual(iface.config.description,
                             f'neutron-{m_pc.current[api.ID]}')
            self.assertEqual(iface.ethernet.switched_vlan.config.operation,
                             nc_op.REPLACE.value)
            self.assertEqual(
                iface.ethernet.switched_vlan.config.interface_mode,
                constants.VLAN_MODE_ACCESS)
            self.assertEqual(
                iface.ethernet.switched_vlan.config.access_vlan,
                segment[api.SEGMENTATION_ID])

    def test_create_port_flat(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_FLAT)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id)
        m_pc.network = m_nc
        segment = {
            api.ID: uuidutils.generate_uuid(),
            api.PHYSICAL_NETWORK:
                m_nc.current['provider:physical_network'],
            api.NETWORK_TYPE: m_nc.current['provider:network_type']}
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.create_port(m_pc, segment, links)
        self.driver.client.edit_config.assert_called_once()
        call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = call_args_list[0][0][0]
        for iface in ifaces:
            self.assertEqual(iface.name, links[0]['port_id'])
            self.assertEqual(iface.config.enabled,
                             m_pc.current['admin_state_up'])
            self.assertEqual(iface.config.mtu, m_nc.current[api.MTU])
            self.assertEqual(iface.config.description,
                             f'neutron-{m_pc.current[api.ID]}')
            self.assertIsNone(iface.ethernet)

    def test_update_port(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15,
            mtu=9000)
        m_nc.original = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15,
            mtu=1500)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            admin_state_up=False)
        m_pc.original = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            admin_state_up=True)
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.update_port(m_pc, links)
        self.driver.client.edit_config.assert_called_once()
        call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = call_args_list[0][0][0]
        for iface in ifaces:
            self.assertEqual(iface.name, links[0]['port_id'])
            self.assertEqual(iface.config.enabled,
                             m_pc.current['admin_state_up'])
            self.assertEqual(iface.config.mtu, m_nc.current[api.MTU])
            self.assertIsNone(iface.ethernet)

    def test_update_port_no_supported_attrib_changed(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        m_nc.original = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            name='current')
        m_pc.original = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            name='original')
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.update_port(m_pc, links)
        self.driver.client.edit_config.assert_not_called()

    def test_delete_port_vlan(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id)
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.delete_port(m_pc, links)
        self.driver.client.edit_config.assert_called_once()
        call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = call_args_list[0][0][0]
        for iface in ifaces:
            self.assertEqual(iface.name, links[0]['port_id'])
            self.assertEqual(iface.config.operation, nc_op.REMOVE.value)
            self.assertEqual(iface.config.description, '')
            self.assertFalse(iface.config.enabled)
            self.assertEqual(iface.config.mtu, 0)
            self.assertEqual(iface.ethernet.switched_vlan.config.operation,
                             nc_op.REMOVE.value)

    def test_delete_port_flat(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_FLAT)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id)
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.delete_port(m_pc, links)
        self.driver.client.edit_config.assert_called_once()
        call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = call_args_list[0][0][0]
        for iface in ifaces:
            self.assertEqual(iface.name, links[0]['port_id'])
            self.assertEqual(iface.config.operation, nc_op.REMOVE.value)
            self.assertEqual(iface.config.description, '')
            self.assertFalse(iface.config.enabled)
            self.assertEqual(iface.config.mtu, 0)
            self.assertIsNone(iface.ethernet)

    def test_create_lacp_port_flat(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        binding_profile = {
            constants.LOCAL_LINK_INFO: [
                {'port_id': 'foo1/1', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'},
                {'port_id': 'foo1/2', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'}],
            constants.LOCAL_GROUP_INFO: {
                'id': uuidutils.generate_uuid(),
                'name': 'PortGroup1',
                'bond_mode': '802.3ad',
                'bond_properties': {
                    constants.LACP_INTERVAL: 'fast',
                    constants.LACP_MIN_LINKS: 2}
            }
        }
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_FLAT)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            binding_profile=binding_profile)
        m_pc.network = m_nc
        segment = {
            api.ID: uuidutils.generate_uuid(),
            api.PHYSICAL_NETWORK:
                m_nc.current['provider:physical_network'],
            api.NETWORK_TYPE: m_nc.current['provider:network_type']}
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.create_port(m_pc, segment, links)
        self.driver.client.edit_config.assert_called_once_with(
            [mock.ANY, mock.ANY], deferred_allocations=True)
        call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = list(call_args_list[0][0][0][0])
        _lacp = call_args_list[0][0][0][1]
        lacp_iface = list(_lacp.interfaces)[0]
        if_link_a = ifaces[0]
        if_link_b = ifaces[1]
        if_agg = ifaces[2]
        assert isinstance(if_link_a, interfaces.InterfaceEthernet)
        assert isinstance(if_link_b, interfaces.InterfaceEthernet)
        assert isinstance(if_agg, interfaces.InterfaceAggregate)
        assert isinstance(lacp_iface, lacp.LACPInterface)
        self.assertEqual(if_link_a.name, 'foo1/1')
        self.assertEqual(if_link_b.name, 'foo1/2')
        self.assertEqual(if_agg.name, openconfig.DEFERRED)
        for iface in (if_link_a, if_link_b):
            self.assertEqual(iface.config.operation, nc_op.MERGE.value)
            self.assertEqual(iface.config.enabled,
                             m_pc.current['admin_state_up'])
            self.assertEqual(iface.config.mtu, m_nc.current['mtu'])
            self.assertEqual(iface.config.description,
                             f'neutron-{m_pc.current[api.ID]}')
            self.assertEqual(iface.ethernet.config.aggregate_id,
                             openconfig.DEFERRED)
        self.assertEqual(if_agg.name, openconfig.DEFERRED)
        self.assertEqual(if_agg.config.name, openconfig.DEFERRED)
        self.assertEqual(if_agg.config.operation, nc_op.MERGE.value)
        self.assertEqual(if_agg.config.description,
                         f'neutron-{m_pc.current[api.ID]}')
        self.assertEqual(if_agg.aggregation.config.lag_type,
                         constants.LAG_TYPE_LACP)
        self.assertEqual(if_agg.aggregation.config.min_links,
                         binding_profile[
                             constants.LOCAL_GROUP_INFO]['bond_properties'][
                             constants.LACP_MIN_LINKS])
        self.assertIsNone(if_agg.aggregation.switched_vlan)
        self.assertEqual(lacp_iface.name, openconfig.DEFERRED)
        self.assertEqual(lacp_iface.operation, nc_op.REPLACE.value)
        self.assertEqual(lacp_iface.config.name, openconfig.DEFERRED)
        self.assertEqual(lacp_iface.config.interval,
                         constants.LACP_PERIOD_FAST)

    def test_create_lacp_port_vlan(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        binding_profile = {
            constants.LOCAL_LINK_INFO: [
                {'port_id': 'foo1/1', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'},
                {'port_id': 'foo1/2', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'}],
            constants.LOCAL_GROUP_INFO: {
                'id': uuidutils.generate_uuid(),
                'name': 'PortGroup1',
                'bond_mode': '802.3ad',
                'bond_properties': {
                    constants.LACP_INTERVAL: 'fast',
                    constants.LACP_MIN_LINKS: 2}
            }
        }
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=40)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            binding_profile=binding_profile)
        m_pc.network = m_nc
        segment = {
            api.ID: uuidutils.generate_uuid(),
            api.PHYSICAL_NETWORK:
                m_nc.current['provider:physical_network'],
            api.NETWORK_TYPE: m_nc.current['provider:network_type'],
            api.SEGMENTATION_ID: m_nc.current['provider:segmentation_id']}
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.create_port(m_pc, segment, links)
        self.driver.client.edit_config.assert_called_once_with(
            [mock.ANY, mock.ANY], deferred_allocations=True)
        call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = list(call_args_list[0][0][0][0])
        _lacp = call_args_list[0][0][0][1]
        lacp_iface = list(_lacp.interfaces)[0]
        if_link_a = ifaces[0]
        if_link_b = ifaces[1]
        if_agg = ifaces[2]
        assert isinstance(if_link_a, interfaces.InterfaceEthernet)
        assert isinstance(if_link_b, interfaces.InterfaceEthernet)
        assert isinstance(if_agg, interfaces.InterfaceAggregate)
        assert isinstance(lacp_iface, lacp.LACPInterface)
        self.assertEqual(if_link_a.name, 'foo1/1')
        self.assertEqual(if_link_b.name, 'foo1/2')
        self.assertEqual(if_agg.name, openconfig.DEFERRED)
        for iface in (if_link_a, if_link_b):
            self.assertEqual(iface.config.operation, nc_op.MERGE.value)
            self.assertEqual(iface.config.enabled,
                             m_pc.current['admin_state_up'])
            self.assertEqual(iface.config.mtu, m_nc.current['mtu'])
            self.assertEqual(iface.config.description,
                             f'neutron-{m_pc.current[api.ID]}')
            self.assertEqual(iface.ethernet.config.aggregate_id,
                             openconfig.DEFERRED)
        self.assertEqual(if_agg.name, openconfig.DEFERRED)
        self.assertEqual(if_agg.config.name, openconfig.DEFERRED)
        self.assertEqual(if_agg.config.operation, nc_op.MERGE.value)
        self.assertEqual(if_agg.config.description,
                         f'neutron-{m_pc.current[api.ID]}')
        self.assertEqual(if_agg.aggregation.config.lag_type,
                         constants.LAG_TYPE_LACP)
        self.assertEqual(if_agg.aggregation.config.min_links,
                         binding_profile[
                             constants.LOCAL_GROUP_INFO]['bond_properties'][
                             constants.LACP_MIN_LINKS])
        self.assertEqual(if_agg.aggregation.switched_vlan.config.operation,
                         nc_op.REPLACE.value)
        self.assertEqual(
            if_agg.aggregation.switched_vlan.config.interface_mode,
            constants.VLAN_MODE_ACCESS)
        self.assertEqual(if_agg.aggregation.switched_vlan.config.access_vlan,
                         segment[api.SEGMENTATION_ID])
        self.assertEqual(lacp_iface.name, openconfig.DEFERRED)
        self.assertEqual(lacp_iface.operation, nc_op.REPLACE.value)
        self.assertEqual(lacp_iface.config.name, openconfig.DEFERRED)
        self.assertEqual(lacp_iface.config.interval,
                         constants.LACP_PERIOD_FAST)

    def test_update_lacp_port(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        binding_profile = {
            constants.LOCAL_LINK_INFO: [
                {'port_id': 'foo1/1', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'},
                {'port_id': 'foo1/2', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'}],
            constants.LOCAL_GROUP_INFO: {
                'id': uuidutils.generate_uuid(),
                'name': 'PortGroup1',
                'bond_mode': '802.3ad',
                'bond_properties': {
                    constants.LACP_INTERVAL: 'fast',
                    constants.LACP_MIN_LINKS: 2}
            }
        }
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15,
            mtu=9000)
        m_nc.original = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15,
            mtu=1500)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            admin_state_up=False, binding_profile=binding_profile)
        m_pc.original = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            admin_state_up=True, binding_profile=binding_profile)
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.client.get.return_value = XML_IFACES_AGGREDATE_ID
        self.driver.update_port(m_pc, links)
        self.driver.client.get.assert_called_once()
        self.driver.client.edit_config.assert_called_once()
        # Validate the query to get aggregate id
        query_call_args = self.driver.client.get.call_args
        ifaces = query_call_args[1]['query']
        iface_a = list(ifaces)[0]
        iface_b = list(ifaces)[1]
        self.assertEqual(iface_a.name, 'foo1/1')
        self.assertEqual(iface_b.name, 'foo1/2')
        self.assertIsNone(iface_a.config)
        self.assertIsNone(iface_a.ethernet)
        self.assertIsNone(iface_b.config)
        self.assertIsNone(iface_b.ethernet)
        # Validate the edit config call
        edit_call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = list(edit_call_args_list[0][0][0])
        if_link_a = ifaces[0]
        if_link_b = ifaces[1]
        if_agg = ifaces[2]
        assert isinstance(if_link_a, interfaces.InterfaceEthernet)
        assert isinstance(if_link_b, interfaces.InterfaceEthernet)
        assert isinstance(if_agg, interfaces.InterfaceAggregate)
        self.assertEqual(if_link_a.name, 'foo1/1')
        self.assertEqual(if_link_b.name, 'foo1/2')
        self.assertEqual(if_agg.name, 'Po10')
        for iface in (if_link_a, if_link_b):
            self.assertEqual(iface.config.operation, nc_op.MERGE.value)
            self.assertEqual(False, iface.config.enabled)
            self.assertEqual(9000, iface.config.mtu)
        self.assertEqual(if_agg.name, 'Po10')
        self.assertEqual(if_agg.operation, nc_op.MERGE.value)
        self.assertEqual(False, if_agg.config.enabled)

    def test_delete_lacp_port_flat(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        binding_profile = {
            constants.LOCAL_LINK_INFO: [
                {'port_id': 'foo1/1', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'},
                {'port_id': 'foo1/2', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'}],
            constants.LOCAL_GROUP_INFO: {
                'id': uuidutils.generate_uuid(),
                'name': 'PortGroup1',
                'bond_mode': '802.3ad',
                'bond_properties': {
                    constants.LACP_INTERVAL: 'fast',
                    constants.LACP_MIN_LINKS: 2}
            }
        }
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_FLAT)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            binding_profile=binding_profile)
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.client.get.return_value = XML_IFACES_AGGREDATE_ID
        self.driver.delete_port(m_pc, links)
        self.driver.client.get.assert_called_once_with(query=mock.ANY)
        self.driver.client.edit_config.assert_called_once_with([mock.ANY,
                                                                mock.ANY])
        # Validate the query to get aggregate id
        query_call_args = self.driver.client.get.call_args
        ifaces = query_call_args[1]['query']
        iface_a = list(ifaces)[0]
        iface_b = list(ifaces)[1]
        self.assertEqual(iface_a.name, 'foo1/1')
        self.assertEqual(iface_b.name, 'foo1/2')
        self.assertIsNone(iface_a.config)
        self.assertIsNone(iface_a.ethernet)
        self.assertIsNone(iface_b.config)
        self.assertIsNone(iface_b.ethernet)
        # Validate the edit config call
        edit_call_args_list = self.driver.client.edit_config.call_args_list
        _lacp = edit_call_args_list[0][0][0][0]
        ifaces = list(edit_call_args_list[0][0][0][1])
        lacp_iface = list(_lacp.interfaces)[0]
        if_link_a = ifaces[0]
        if_link_b = ifaces[1]
        if_agg = ifaces[2]
        assert isinstance(if_link_a, interfaces.InterfaceEthernet)
        assert isinstance(if_link_b, interfaces.InterfaceEthernet)
        assert isinstance(if_agg, interfaces.InterfaceAggregate)
        assert isinstance(lacp_iface, lacp.LACPInterface)
        self.assertEqual(if_link_a.name, 'foo1/1')
        self.assertEqual(if_link_b.name, 'foo1/2')
        self.assertEqual(if_agg.name, 'Po10')
        for iface in (if_link_a, if_link_b):
            self.assertEqual(iface.config.operation, nc_op.REMOVE.value)
            self.assertEqual(iface.config.enabled, False)
            self.assertEqual(iface.config.mtu, 0)
            self.assertEqual(iface.config.description, '')
            self.assertIsNone(iface.ethernet.switched_vlan)
            self.assertIsNone(iface.ethernet.config.aggregate_id)
            self.assertEqual(iface.ethernet.config.operation,
                             nc_op.REMOVE.value)
        self.assertEqual(if_agg.name, 'Po10')
        self.assertEqual(if_agg.operation, nc_op.REMOVE.value)
        self.assertIsNone(if_agg.config)
        self.assertIsNone(if_agg.aggregation)
        self.assertEqual(lacp_iface.name, 'Po10')
        self.assertEqual(lacp_iface.operation, nc_op.REMOVE.value)
        self.assertIsNone(lacp_iface.config)

    def test_delete_lacp_port_vlan(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        binding_profile = {
            constants.LOCAL_LINK_INFO: [
                {'port_id': 'foo1/1', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'},
                {'port_id': 'foo1/2', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'}],
            constants.LOCAL_GROUP_INFO: {
                'id': uuidutils.generate_uuid(),
                'name': 'PortGroup1',
                'bond_mode': '802.3ad',
                'bond_properties': {
                    constants.LACP_INTERVAL: 'fast',
                    constants.LACP_MIN_LINKS: 2}
            }
        }
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=40)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            binding_profile=binding_profile)
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.client.get.return_value = XML_IFACES_AGGREDATE_ID
        self.driver.delete_port(m_pc, links)
        self.driver.client.get.assert_called_once_with(query=mock.ANY)
        self.driver.client.edit_config.assert_called_once_with([mock.ANY,
                                                                mock.ANY])
        # Validate the query to get aggregate id
        query_call_args = self.driver.client.get.call_args
        ifaces = query_call_args[1]['query']
        iface_a = list(ifaces)[0]
        iface_b = list(ifaces)[1]
        self.assertEqual(iface_a.name, 'foo1/1')
        self.assertEqual(iface_b.name, 'foo1/2')
        self.assertIsNone(iface_a.config)
        self.assertIsNone(iface_a.ethernet)
        self.assertIsNone(iface_b.config)
        self.assertIsNone(iface_b.ethernet)
        # Validate the edit config call
        edit_call_args_list = self.driver.client.edit_config.call_args_list
        _lacp = edit_call_args_list[0][0][0][0]
        ifaces = list(edit_call_args_list[0][0][0][1])
        lacp_iface = list(_lacp.interfaces)[0]
        if_link_a = ifaces[0]
        if_link_b = ifaces[1]
        if_agg = ifaces[2]
        assert isinstance(if_link_a, interfaces.InterfaceEthernet)
        assert isinstance(if_link_b, interfaces.InterfaceEthernet)
        assert isinstance(if_agg, interfaces.InterfaceAggregate)
        assert isinstance(lacp_iface, lacp.LACPInterface)
        self.assertEqual(if_link_a.name, 'foo1/1')
        self.assertEqual(if_link_b.name, 'foo1/2')
        self.assertEqual(if_agg.name, 'Po10')
        for iface in (if_link_a, if_link_b):
            self.assertEqual(iface.config.operation, nc_op.REMOVE.value)
            self.assertEqual(iface.config.enabled, False)
            self.assertEqual(iface.config.mtu, 0)
            self.assertEqual(iface.config.description, '')
            self.assertEqual(iface.ethernet.switched_vlan.config.operation,
                             nc_op.REMOVE.value)
            self.assertIsNone(iface.ethernet.config.aggregate_id)
        self.assertEqual(if_agg.name, 'Po10')
        self.assertEqual(if_agg.operation, nc_op.REMOVE.value)
        self.assertIsNone(if_agg.config)
        self.assertIsNone(if_agg.aggregation)
        self.assertEqual(lacp_iface.name, 'Po10')
        self.assertEqual(lacp_iface.operation, nc_op.REMOVE.value)
        self.assertIsNone(lacp_iface.config)

    def test_create_pre_configured_aggregate_port_vlan(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        binding_profile = {
            constants.LOCAL_LINK_INFO: [
                {'port_id': 'foo1/1', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'},
                {'port_id': 'foo1/2', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'}],
            constants.LOCAL_GROUP_INFO: {
                'id': uuidutils.generate_uuid(),
                'name': 'PortGroup1',
                'bond_mode': 'balance-rr',
            }
        }
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=40)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            binding_profile=binding_profile)
        m_pc.network = m_nc
        segment = {
            api.ID: uuidutils.generate_uuid(),
            api.PHYSICAL_NETWORK:
                m_nc.current['provider:physical_network'],
            api.NETWORK_TYPE: m_nc.current['provider:network_type'],
            api.SEGMENTATION_ID: m_nc.current['provider:segmentation_id']}
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.client.get.return_value = XML_IFACES_AGGREDATE_ID
        self.driver.create_port(m_pc, segment, links)
        self.driver.client.get.assert_called_once()
        self.driver.client.edit_config.assert_called_once()
        # Validate the query to get aggregate id
        query_call_args = self.driver.client.get.call_args
        ifaces = query_call_args[1]['query']
        iface_a = list(ifaces)[0]
        iface_b = list(ifaces)[1]
        self.assertEqual(iface_a.name, 'foo1/1')
        self.assertEqual(iface_b.name, 'foo1/2')
        self.assertIsNone(iface_a.config)
        self.assertIsNone(iface_a.ethernet)
        self.assertIsNone(iface_b.config)
        self.assertIsNone(iface_b.ethernet)
        # Validate the edit config call
        edit_call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = list(edit_call_args_list[0][0][0])
        if_agg = ifaces[0]
        assert isinstance(if_agg, interfaces.InterfaceAggregate)
        self.assertEqual(if_agg.name, 'Po10')
        self.assertEqual(if_agg.operation, nc_op.MERGE.value)
        self.assertEqual(True, if_agg.config.enabled)
        self.assertEqual(if_agg.aggregation.switched_vlan.config.operation,
                         nc_op.REPLACE.value)
        self.assertEqual(
            if_agg.aggregation.switched_vlan.config.interface_mode,
            constants.VLAN_MODE_ACCESS)
        self.assertEqual(if_agg.aggregation.switched_vlan.config.access_vlan,
                         segment[api.SEGMENTATION_ID])

    def test_update_pre_configured_aggregate_port_vlan(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        binding_profile = {
            constants.LOCAL_LINK_INFO: [
                {'port_id': 'foo1/1', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'},
                {'port_id': 'foo1/2', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'}],
            constants.LOCAL_GROUP_INFO: {
                'id': uuidutils.generate_uuid(),
                'name': 'PortGroup1',
                'bond_mode': 'balance-rr',
            }
        }
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        m_nc.original = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            admin_state_up=False, binding_profile=binding_profile)
        m_pc.original = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            admin_state_up=True, binding_profile=binding_profile)
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.client.get.return_value = XML_IFACES_AGGREDATE_ID
        self.driver.update_port(m_pc, links)
        self.driver.client.get.assert_called_once()
        self.driver.client.edit_config.assert_called_once()
        # Validate the edit config call
        edit_call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = list(edit_call_args_list[0][0][0])
        if_agg = ifaces[0]
        assert isinstance(if_agg, interfaces.InterfaceAggregate)
        self.assertEqual(if_agg.name, 'Po10')
        self.assertEqual(if_agg.operation, nc_op.MERGE.value)
        self.assertEqual(False, if_agg.config.enabled)

    def test_delete_pre_configured_aggregate_port_vlan(self):
        tenant_id = uuidutils.generate_uuid()
        network_id = uuidutils.generate_uuid()
        project_id = uuidutils.generate_uuid()
        m_nc = mock.create_autospec(driver_context.NetworkContext)
        m_pc = mock.create_autospec(driver_context.PortContext)
        binding_profile = {
            constants.LOCAL_LINK_INFO: [
                {'port_id': 'foo1/1', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'},
                {'port_id': 'foo1/2', 'switch_id': 'aa:bb:cc:dd:ee:ff',
                 'switch_info': 'foo'}],
            constants.LOCAL_GROUP_INFO: {
                'id': uuidutils.generate_uuid(),
                'name': 'PortGroup1',
                'bond_mode': 'balance-rr',
            }
        }
        m_nc.current = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        m_nc.original = ml2_utils.get_test_network(
            id=network_id, tenant_id=tenant_id, project_id=project_id,
            network_type=n_const.TYPE_VLAN, segmentation_id=15)
        m_pc.current = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            admin_state_up=False, binding_profile=binding_profile)
        m_pc.original = ml2_utils.get_test_port(
            network_id=network_id, tenant_id=tenant_id, project_id=project_id,
            admin_state_up=True, binding_profile=binding_profile)
        m_pc.network = m_nc
        links = m_pc.current['binding:profile'][constants.LOCAL_LINK_INFO]
        self.driver.client.get.return_value = XML_IFACES_AGGREDATE_ID
        self.driver.delete_port(m_pc, links)
        self.driver.client.get.assert_called_once()
        self.driver.client.edit_config.assert_called_once()
        # Validate the edit config call
        edit_call_args_list = self.driver.client.edit_config.call_args_list
        ifaces = list(edit_call_args_list[0][0][0])
        if_agg = ifaces[0]
        assert isinstance(if_agg, interfaces.InterfaceAggregate)
        self.assertEqual(if_agg.name, 'Po10')
        self.assertEqual(if_agg.operation, nc_op.MERGE.value)
        self.assertEqual(False, if_agg.config.enabled)
        self.assertEqual(if_agg.aggregation.switched_vlan.config.operation,
                         nc_op.REMOVE.value)



# Copyright 2017 Cisco Systems, Inc.
# All Rights Reserved
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import enum

BAREMETAL_AGENT_TYPE = "Baremetal Node"
BAREMETAL_BINARY = 'ironic-neutron-agent'

LOCAL_LINK_INFO = 'local_link_information'
LOCAL_GROUP_INFO = 'local_group_information'

IFACE_TYPE_ETHERNET = 'ethernet'
IFACE_TYPE_AGGREGATE = 'aggregate'
IFACE_TYPE_BASE = 'base'

LAG_TYPE_LACP = 'LACP'
LAG_TYPE_SATIC = 'SATIC'

LACP_TIMEOUT_LONG = 'LONG'
LACP_TIMEOUT_SHORT = 'SHORT'
LACP_PERIOD_FAST = 'FAST'
LACP_PERIOD_SLOW = 'SLOW'
LACP_ACTIVITY_ACTIVE = 'ACTIVE'
LACP_ACTIVITY_PASSIVE = 'PASSIVE'
LACP_MIN_LINKS = 'bond_min_links'
LACP_INTERVAL = 'bond_lacp_rate'

# These bond modes require switch configuration the plugin cannot create.
PRE_CONF_ONLY_BOND_MODES = {'balance-rr', '0',
                            'balance-xor', '2',
                            'broadcast', '3'}
LACP_BOND_MODES = {'802.3ad', '4'}
NON_SWITCH_BOND_MODES = {'active-backup', '1',
                         'balance-tlb', '5',
                         'balance-alb', '6'}

VLAN_ACTIVE = 'ACTIVE'
VLAN_SUSPENDED = 'SUSPENDED'
VLAN_MODE_TRUNK = 'TRUNK'
VLAN_MODE_ACCESS = 'ACCESS'
VLAN_RANGE = range(1, 4094)

PORT_ID = 'port_id'
SWITCH_ID = 'switch_id'
SWITCH_INFO = 'switch_info'


class NetconfEditConfigOperation(enum.Enum):
    """RFC 6241 - <edit-config> operation attribute

    The "operation" attribute has one of the following values:
         merge:  The configuration data identified by the element
            containing this attribute is merged with the configuration
            at the corresponding level in the configuration datastore
            identified by the <target> parameter.  This is the default
            behavior.
         replace:  The configuration data identified by the element
            containing this attribute replaces any related configuration
            in the configuration datastore identified by the <target>
            parameter.  If no such configuration data exists in the
            configuration datastore, it is created.  Unlike a
            <copy-config> operation, which replaces the entire target
            configuration, only the configuration actually present in
            the <config> parameter is affected.
         create:  The configuration data identified by the element
            containing this attribute is added to the configuration if
            and only if the configuration data does not already exist in
            the configuration datastore.  If the configuration data
            exists, an <rpc-error> element is returned with an
            <error-tag> value of "data-exists".
         delete:  The configuration data identified by the element
            containing this attribute is deleted from the configuration
            if and only if the configuration data currently exists in
            the configuration datastore.  If the configuration data does
            not exist, an <rpc-error> element is returned with an
            <error-tag> value of "data-missing".
         remove:  The configuration data identified by the element
            containing this attribute is deleted from the configuration
            if the configuration data currently exists in the
            configuration datastore.  If the configuration data does not
            exist, the "remove" operation is silently ignored by the
            server.
    """
    MERGE = 'merge'
    REPLACE = 'replace'
    CREATE = 'create'
    DELETE = 'delete'
    REMOVE = 'remove'


CFG_ELEMENT = 'config'

IANA_NETCONF_CAPABILITIES = {
    # [RFC4741][RFC6241]
    ':base:1.0':
        'urn:ietf:params:netconf:base:1.0',
    # [RFC4741]
    ':confirmed-commit':
        'urn:ietf:params:netconf:capability:confirmed-commit:1.0',
    ':validate':
        'urn:ietf:params:netconf:capability:validate:1.0',
    # [RFC6241]
    ':base:1.1':
        'urn:ietf:params:netconf:base:1.1',
    ':writable-running':
        'urn:ietf:params:netconf:capability:writable-running:1.0',
    ':candidate':
        'urn:ietf:params:netconf:capability:candidate:1.0',
    ':confirmed-commit:1.1':
        'urn:ietf:params:netconf:capability:confirmed-commit:1.1',
    ':rollback-on-error':
        'urn:ietf:params:netconf:capability:rollback-on-error:1.0',
    ':validate:1.1':
        'urn:ietf:params:netconf:capability:validate:1.1',
    ':startup':
        'urn:ietf:params:netconf:capability:startup:1.0',
}



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from xml.etree import ElementTree

from networking_baremetal import common
from networking_baremetal import constants
from networking_baremetal.openconfig.vlan import vlan


class InterfacesEthernetConfig:
    """OpenConfig interface ethernet configuration"""

    NAMESPACE = 'http://openconfig.net/yang/interfaces'
    PARENT = 'interface'
    TAG = 'config'

    def __init__(self, operation=constants.NetconfEditConfigOperation.MERGE):
        self.operation = operation
        self._aggregate_id = None
        self._aggregate_id_namespace = (
            'http://openconfig.net/yang/interfaces/aggregate')

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value if self._operation else None

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation attribute.'
                            .format(type(value)))

    @operation.deleter
    def operation(self):
        self._operation = None

    @property
    def aggregate_id(self):
        """Logical aggregate interface for interface"""
        return self._aggregate_id

    @aggregate_id.setter
    def aggregate_id(self, value: str):
        """Set logical aggregate interface for interface"""
        if not isinstance(value, str):
            raise TypeError('aggregate_id must be string, got {}'
                            .format(type(value)))
        self._aggregate_id = value

    @aggregate_id.deleter
    def aggregate_id(self):
        self._aggregate_id = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        element = ElementTree.Element(self.TAG)
        if self.operation:
            element.set('operation', self.operation)
        if self.aggregate_id is not None:
            common.txt_subelement(element, 'aggregate-id', self.aggregate_id,
                                  xmlns=self._aggregate_id_namespace)
        return element


class InterfacesEthernet:
    """Ethernet configuration and state"""
    NAMESPACE = 'http://openconfig.net/yang/interfaces/ethernet'
    PARENT = 'interface'
    TAG = 'ethernet'

    def __init__(self):
        self._switched_vlan = vlan.VlanSwitchedVlan()
        self._config = InterfacesEthernetConfig()

    @property
    def switched_vlan(self):
        return self._switched_vlan

    @switched_vlan.setter
    def switched_vlan(self, value):
        if not isinstance(value, vlan.VlanSwitchedVlan):
            raise TypeError('switched_vlan must be VlanSwitchedVlan, got {}'
                            .format(type(value)))
        self._switched_vlan = value

    @switched_vlan.deleter
    def switched_vlan(self):
        self._switched_vlan = None

    @property
    def config(self):
        """Configuration parameters for interface"""
        return self._config

    @config.setter
    def config(self, value):
        if not isinstance(value, InterfacesEthernetConfig):
            raise TypeError('config must be InterfacesEthernetConfig, got {}'
                            .format(type(value)))
        self._config = value

    @config.deleter
    def config(self):
        self._config = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('xmlns', self.NAMESPACE)
        if self.config:
            elem.append(self.config.to_xml_element())
        if self.switched_vlan:
            elem.append(self.switched_vlan.to_xml_element())
        return elem



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from xml.etree import ElementTree

from networking_baremetal import common
from networking_baremetal import constants
from networking_baremetal.openconfig.interfaces import types
from networking_baremetal.openconfig.vlan import vlan


class InterfacesAggregationConfig:

    NAMESPACE = 'http://openconfig.net/yang/interfaces/aggregate'
    PARENT = 'aggregation'
    TAG = 'config'

    def __init__(self,
                 operation: str = constants.NetconfEditConfigOperation.MERGE):
        self.operation = operation
        self._lag_type = None
        self._min_links = None

    @property
    def operation(self):
        """RFC 6241 - <edit-config> operation attribute"""
        return self._operation.value if self._operation else None

    @operation.setter
    def operation(self, value):
        """RFC 6241 - <edit-config> operation attribute"""
        if isinstance(value, constants.NetconfEditConfigOperation):
            self._operation = value
        elif isinstance(value, str):
            self._operation = constants.NetconfEditConfigOperation(value)
        else:
            raise TypeError('Invalid type {} for config operation attribute.'
                            .format(type(value)))

    @operation.deleter
    def operation(self):
        self._operation = None

    @property
    def lag_type(self):
        return self._lag_type.value if self._lag_type else None

    @lag_type.setter
    def lag_type(self, value: str):
        """the type of LAG, i.e., how it is configured / maintained"""
        self._lag_type = types.AggregationType(value)

    @lag_type.deleter
    def lag_type(self):
        self._lag_type = None

    @property
    def min_links(self):
        return self._min_links

    @min_links.setter
    def min_links(self, value: int):
        self._min_links = value

    @min_links.deleter
    def min_links(self):
        self._min_links = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        if self.operation:
            elem.set('operation', self.operation)
        if self.lag_type is not None:
            common.txt_subelement(elem, 'lag-type', self.lag_type)
        if self.min_links is not None:
            common.txt_subelement(elem, 'min-links', str(self.min_links))
        return elem


class InterfacesAggregation:
    """Options for logical interfaces representing aggregates"""

    NAMESPACE = 'http://openconfig.net/yang/interfaces/aggregate'
    PARENT = 'interface'
    TAG = 'aggregation'

    def __init__(self):
        self._switched_vlan = vlan.VlanSwitchedVlan()
        self._config = InterfacesAggregationConfig()

    @property
    def switched_vlan(self):
        return self._switched_vlan

    @switched_vlan.setter
    def switched_vlan(self, value):
        if not isinstance(value, vlan.VlanSwitchedVlan):
            raise TypeError('switched_vlan must be '
                            'OpenConfigVlanSwitchedVlan, got {}'
                            .format(type(value)))
        self._switched_vlan = value

    @switched_vlan.deleter
    def switched_vlan(self):
        self._switched_vlan = None

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, value):
        if not isinstance(value, InterfacesAggregationConfig):
            raise TypeError('config must be InterfacesAggregationConfig, got '
                            '{}'.format(type(value)))
        self._config = value

    @config.deleter
    def config(self):
        self._config = None

    def to_xml_element(self):
        """Create XML Element

        :return: ElementTree Element with SubElements
        """
        elem = ElementTree.Element(self.TAG)
        elem.set('xmlns', self.NAMESPACE)
        if self.config:
            elem.append(self.config.to_xml_element())
        if self.switched_vlan:
            elem.append(self.switched_vlan.to_xml_element())
        return elem
